from typing import List, Optional
from attrs import evolve
import torch
from experimaestro import Param
from xpmir.letor.records import TopicRecord, DocumentRecord
from xpmir.neural import DualRepresentationScorer, QueriesRep, DocsRep
from xpmir.utils.utils import easylog, foreach
from xpmir.text.encoders import TextEncoderBase
from xpmir.learning.context import Loss, TrainerContext, TrainingHook
from xpmir.learning.metrics import ScalarMetric

logger = easylog()


class DualVectorListener(TrainingHook):
    """Listener called with the (vectorial) representation of queries and
    documents

    The hook is called just after the computation of documents and queries
    representations.

    This can be used for logging purposes, but more importantly, to add
    regularization losses such as the :class:`FlopsRegularizer` regularizer.
    """

    def __call__(
        self, context: TrainerContext, queries: torch.Tensor, documents: torch.Tensor
    ):
        """Hook handler

        Args:
            context (TrainerContext): The training context
            queries (torch.Tensor): The query vectors
            documents (torch.Tensor): The document vectors

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError(f"__call__ in {self.__class__}")


class DualVectorScorer(DualRepresentationScorer[QueriesRep, DocsRep]):
    """A scorer based on dual vectorial representations"""

    encoder: Param[TextEncoderBase]
    """The document (and potentially query) encoder"""

    query_encoder: Param[Optional[TextEncoderBase]]
    """The query encoder (optional, if not defined uses the query_encoder)"""

    def __initialize__(self, options):
        self.encoder.initialize(options)
        if self.query_encoder:
            self.query_encoder.initialize(options)

    @property
    def _query_encoder(self):
        return self.query_encoder or self.encoder

    def __validate__(self):
        super().__validate__()
        assert not self.encoder.static(), "The vocabulary should be learnable"


class Dense(DualVectorScorer[QueriesRep, DocsRep]):
    """A scorer based on a pair of (query, document) dense vectors"""

    def score_product(self, queries, documents, info: Optional[TrainerContext] = None):
        if info is not None:
            foreach(
                info.hooks(DualVectorListener),
                lambda hook: hook(info, queries, documents),
            )

        return queries.value @ documents.value.T

    def score_pairs(self, queries, documents, info: Optional[TrainerContext] = None):
        scores = (
            (queries.value.unsqueeze(1) @ documents.value.unsqueeze(2))
            .squeeze(-1)
            .squeeze(-1)
        )

        # Apply the dual vector hook
        if info is not None:
            foreach(
                info.hooks(DualVectorListener),
                lambda hook: hook(info, queries, documents),
            )
        return scores

    @classmethod
    def from_sentence_transformers(cls, hf_id: str, **kwargs):
        """Creates a dense model from a Sentence transformer

        The list can be found on HuggingFace
        https://huggingface.co/models?library=sentence-transformers

        :param hf_id: The HuggingFace ID
        """
        from xpmir.text.huggingface import SentenceTransformerTextEncoder

        encoder = SentenceTransformerTextEncoder(hf_id)
        return cls(encoder, **kwargs)


class CosineDense(Dense):
    """Dual model based on cosine similarity."""

    def encode_queries(self, records: List[TopicRecord]):
        queries = (self.query_encoder or self.encoder)(records)
        return evolve(
            queries, value=queries.value / queries.value.norm(dim=-1, keepdim=True)
        )

    def encode_documents(self, records: List[DocumentRecord]):
        documents = self.encoder(records)
        return evolve(
            documents,
            value=documents.value / documents.value.norm(dim=-1, keepdim=True),
        )

class DotDense(Dense):
    """Dual model based on inner product."""

    def __validate__(self):
        super().__validate__()
        assert not self.encoder.static(), "The vocabulary should be learnable"

    def encode_queries(self, records: List[TopicRecord]):
        """Encode the different queries"""
        return self._query_encoder(records)

    def encode_documents(self, records: List[DocumentRecord]):
        """Encode the different documents"""
        return self.encoder(records)


class FlopsRegularizer(DualVectorListener):
    r"""The FLOPS regularizer computes

    .. math::

        FLOPS(q,d) = \lambda_q FLOPS(q) + \lambda_d FLOPS(d)

    where

    .. math::
        FLOPS(x) = \left( \frac{1}{d} \sum_{i=1}^d |x_i| \right)^2
    """
    lambda_q: Param[float]
    """Lambda for queries"""

    lambda_d: Param[float]
    """Lambda for documents"""

    def compute(x: torch.Tensor):
        """
        :param x: term vectors (batch size x vocabulary dimension)
        :returns: A couple (vocabulary dimension / FLOPS regularizations)
        """
        # Computes the mean for each term (weights are positive)
        y = x.mean(0)

        # Returns the sum of squared means
        return y, (y * y).sum()

    def __call__(self, info: TrainerContext, queries, documents):
        # queries and documents are length x dimension
        # Assumes that all weights are positive
        assert info.metrics is not None

        queries = queries.value
        documents = documents.value

        # q of shape (dimension), flops_q of shape (1)
        q, flops_q = FlopsRegularizer.compute(queries)
        d, flops_d = FlopsRegularizer.compute(documents)

        flops = self.lambda_d * flops_d + self.lambda_q * flops_q
        info.add_loss(Loss("flops", flops, 1.0))

        info.metrics.add(ScalarMetric("flops", flops.item(), 1))
        info.metrics.add(ScalarMetric("flops_q", flops_q.item(), 1))
        info.metrics.add(ScalarMetric("flops_d", flops_d.item(), 1))

        with torch.no_grad():
            info.metrics.add(
                ScalarMetric(
                    "sparsity_q",
                    torch.count_nonzero(queries).item()
                    / (queries.shape[0] * queries.shape[1]),
                    len(q),
                )
            )
            info.metrics.add(
                ScalarMetric(
                    "sparsity_d",
                    torch.count_nonzero(documents).item()
                    / (documents.shape[0] * documents.shape[1]),
                    len(d),
                )
            )


class ScheduledFlopsRegularizer(FlopsRegularizer):
    """
    The FLOPS regularizer where the lamdba_q and lambda_d varie according to the
    steps. The lambda values goes quadratic before the
    ```lambda_warmup_steps```, and then remains constant
    """

    min_lambda_q: Param[float] = 0
    """Min value for the lambda_q before it increase"""

    min_lambda_d: Param[float] = 0
    """Min value for the lambda_d before it increase"""

    lambda_warmup_steps: Param[int] = 0
    """The warmup steps for the lambda"""

    def quadratic_ratio(self, step):
        if step > self.lambda_warmup_steps:
            return 1
        else:
            return (step / self.lambda_warmup_steps) ** 2

    def __post_init__(self):
        self.initial_lambda_q = self.lambda_q
        self.initial_lambda_d = self.lambda_d

    def __call__(self, info: TrainerContext, queries, documents):
        current_step = info.steps
        self.lambda_q = (
            self.initial_lambda_q - self.min_lambda_q
        ) * self.quadratic_ratio(current_step) + self.min_lambda_q
        self.lambda_d = (
            self.initial_lambda_d - self.min_lambda_d
        ) * self.quadratic_ratio(current_step) + self.min_lambda_d
        FlopsRegularizer.__call__(self, info, queries, documents)



import pytest
import torch
import numpy as np
from pathlib import Path
from datamaestro_text.data.ir import TextItem, IDItem
from experimaestro import ObjectStore
from experimaestro.xpmutils import DirectoryContext
from xpmir.index.sparse import SparseRetriever, SparseRetrieverIndexBuilder
from xpmir.test.utils.utils import SampleDocumentStore, SparseRandomTextEncoder


@pytest.fixture
def context(tmp_path: Path):
    from experimaestro.taskglobals import Env

    Env.taskpath = tmp_path / "task"
    Env.taskpath.mkdir()
    return DirectoryContext(tmp_path)


class SparseIndex:
    def __init__(self, context, ordered_index: bool = False):
        # Build the index
        objects = ObjectStore()
        documents = SampleDocumentStore(num_docs=500)
        self.encoder = SparseRandomTextEncoder(dim=1000, sparsity=0.8)
        builder = SparseRetrieverIndexBuilder(
            encoder=self.encoder,
            documents=documents,
            max_postings=10,
            batch_size=5,
            ordered_index=ordered_index,
        )

        # Build the index
        builder_instance: SparseRetrieverIndexBuilder = builder.instance(
            context=context, objects=objects
        )
        builder_instance.execute()

        self.document_store = builder_instance.documents
        self.x_docs = builder_instance.encoder(
            [d[TextItem].text for d in self.document_store.documents.values()]
        ).value

        # Check index
        self.index = builder.task_outputs(lambda x: x)
        self.index_instance = self.index.instance(objects=objects)
        self.topk = 10


@pytest.fixture(params=[False])
def sparse_index(context, request):
    return SparseIndex(context, ordered_index=request.param)


def test_sparse_indexation(sparse_index: SparseIndex):
    for ix in np.random.choice(np.arange(len(sparse_index.x_docs.T)), 10):
        # for ix in range(len(sparse_index.x_docs.T)):
        x = sparse_index.x_docs[:, ix]

        # nz indices are indices of documents
        nz = torch.nonzero(x)

        if sparse_index.index_instance.ordered:
            sorted_ix = sorted(range(len(nz)), key=lambda jx: -x[nz[jx]])
            nz = nz[sorted_ix]

        sparse_index.index_instance.initialize(False)
        for jx, posting in enumerate(sparse_index.index_instance.index.postings(ix)):
            assert (
                posting.docid == nz[jx]
            ), f"Error for posting {jx} of term {ix} (docid)"
            assert (
                x[nz[jx]] == posting.value
            ), f"Error for posting {jx} of term {ix} (value)"

            jx += 1


@pytest.fixture(params=[False])
def retriever(context, sparse_index: SparseIndex, request: bool):
    retriever = SparseRetriever(
        encoder=sparse_index.encoder,
        topk=sparse_index.topk,
        batchsize=2,
        index=sparse_index.index,
        in_memory=request.param,
    ).instance(context=context)
    retriever.initialize()

    return retriever


def test_sparse_retrieve(sparse_index: SparseIndex, retriever):
    # Computes the score directly
    x_docs = sparse_index.x_docs.type(torch.float32)

    # Choose a few documents
    chosen_ix = np.random.choice(np.arange(len(sparse_index.x_docs)), 10)
    for ix in chosen_ix:
        document = sparse_index.document_store.document_int(ix)

        # Use the retriever
        scoredDocuments = retriever.retrieve(document[TextItem].text)
        # scoredDocuments.sort(reverse=True)
        # scoredDocuments = scoredDocuments[:retriever.topk]

        # Use the pre-computed scores
        scores = x_docs[ix] @ x_docs.T
        sorted = scores.sort(descending=True, stable=True)
        indices = sorted.indices[: retriever.topk]
        expected = list(indices.numpy())

        observed = [int(sd.document[IDItem].id) for sd in scoredDocuments]
        expected_scores = sorted.values[: retriever.topk].numpy()
        observed_scores = np.array([float(sd.score) for sd in scoredDocuments])

        np.testing.assert_allclose(
            expected_scores,
            observed_scores,
            1e-5,
            err_msg=f"{ix} {expected} vs {observed}",
        )
        assert expected == observed


def test_sparse_retrieve_all(retriever):
    """Just verifies that the retriever is coherent with itself when retrieving
    many queries"""
    queries = {
        "q1": "Query 1",
        "q2": "Query 2",
        "q3": "Query 3",
        "q4": "Query 4",
        "q5": "Query 55",
    }
    all_results = retriever.retrieve_all(queries)

    for key, query in queries.items():
        query_results = retriever.retrieve(query)

        observed = [d.document[IDItem].id for d in all_results[key]]
        expected = [d.document[IDItem].id for d in query_results]
        assert observed == expected

        observed_scores = [d.score for d in all_results[key]]
        expected_scores = [d.score for d in query_results]
        np.testing.assert_allclose(
            expected_scores, observed_scores, 1e-5, err_msg=f"{expected} vs {observed}"
        )



from torch import nn
import numpy as np
from experimaestro import Param, Config
import logging

from xpmir.letor.samplers import PairwiseSampler
from xpmir.letor.records import BaseRecords, PairwiseRecords
from xpmir.letor.trainers import TrainerContext, LossTrainer
from xpmir.learning.context import Loss
from xpmir.utils.iter import MultiprocessSerializableIterator
from xpmir.utils.utils import foreach, easylog

logger = easylog()


class PairwiseGenerativeLoss(Config, nn.Module):
    """Generic loss for generative models"""

    NAME = "?"

    weight: Param[float] = 1.0
    """The weight :math:`w` with which the loss is multiplied (useful when
    combining with other ones)"""

    def compute(self, records, context):
        pass

    def process(self, records: BaseRecords, context: TrainerContext):
        value = self.compute(records, context)  # tensor shape [bs]
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"Loss: {value}")
        context.add_loss(Loss(f"pair-{self.NAME}", value, self.weight))


class GenerativeTrainer(LossTrainer):
    loss: Param[PairwiseGenerativeLoss]

    sampler: Param[PairwiseSampler]
    """The pairwise sampler"""

    def initialize(self, random: np.random.RandomState, context: TrainerContext):
        super().initialize(random, context)
        self.loss.initialize()
        foreach(
            context.hooks(PairwiseGenerativeLoss), lambda loss: loss.initialize()
        )  # maybe later we need to change the sampling target, we can use this hook

        self.sampler.initialize(random)
        self.sampler_iter = MultiprocessSerializableIterator(
            self.sampler.pairwise_batch_iter(self.batch_size)
        )

    def train_batch(self, records: PairwiseRecords):
        # do the forward pass to get the gradient value
        self.loss.process(records, self.context)



from typing import Optional
import ir_measures as irm
from experimaestro import Param
from datamaestro_text.data.ir import Measure as BaseMeasure


class Measure(BaseMeasure):
    """Mirrors the ir_measures metric object"""

    identifier: Param[str]
    """main identifier"""

    rel: Param[int] = 1
    """minimum relevance score to be considered relevant (inclusive)"""

    cutoff: Param[Optional[int]]
    """Cutoff value"""

    def __matmul__(self, cutoff):
        return Measure.C(identifier=self.identifier, rel=self.rel, cutoff=int(cutoff))

    def __call__(self):
        measure = irm.parse_measure(self.identifier)
        if self.cutoff is not None:
            measure = measure @ self.cutoff
        return measure

    def __repr__(self):
        return f"{self.identifier}@{self.cutoff}/rel={self.rel}"


AP = Measure.C(identifier="AP")
"""Average precision metric"""

P = Measure.C(identifier="P")
"""Precision at rank"""

RR = Measure.C(identifier="RR")
"""Reciprocical rank"""

nDCG = Measure.C(identifier="nDCG")
"""Normalized Discounted Cumulated Gain"""

R = Measure.C(identifier="R")
"""Recall at rank"""

Success = Measure.C(identifier="Success")
"""1 if a document with at least rel relevance is found in the first cutoff
documents, else 0."""



from experimaestro import Config, Param
from experimaestro.compat import cached_property
import numpy as np
from xpmir.utils.utils import easylog

# flake8: noqa: F401
from ..learning.devices import (
    DEFAULT_DEVICE,
    Device,
    DeviceInformation,
    DistributedDeviceInformation,
)

# flake8: noqa: F401
from ..learning import Random

logger = easylog()



import pickle
import hashlib
from typing import Optional
import numpy as np
import torch
from pathlib import Path
from torch import nn
from experimaestro import cache, Param

from xpmir.letor import Random
from xpmir.text.encoders import TokenizedTexts
from xpmir.learning import ModuleInitOptions, ModuleInitMode

from xpmir.text import TokensEncoder
from datamaestro_text.data.embeddings import WordEmbeddings


class WordvecVocab(TokensEncoder, nn.Module):
    """Word-based pre-trained embeddings

    Args:

        train: Should the word embeddings be re-retrained?
    """

    data: Param[WordEmbeddings]
    learn: Param[bool] = False
    random: Param[Optional[Random]]

    def __post_init__(self):
        # super().__post_init__()
        self._terms, self._weights = self.load()
        self._term2idx = {t: i for i, t in enumerate(self._terms)}
        matrix = self._weights
        self.size = matrix.shape[1]
        matrix = np.concatenate(
            [np.zeros((1, self.size)), matrix]
        )  # add padding record (-1)
        self.embed = nn.Embedding.from_pretrained(
            torch.from_numpy(matrix.astype(np.float32)), freeze=not self.learn
        )

    @property
    def pad_tokenid(self):
        return 0

    def learnable(self):
        return self.learn

    def __validate__(self):
        """Check that values are coherent"""
        if self.learn:
            assert self.random is not None

    @cache("terms.npy")
    def load(self, path: Path):
        path_lst = path.with_suffix(".lst")
        if path.is_file():
            with path_lst.open("rb") as fp:
                return pickle.load(fp), np.load(path)

        terms, weights = self.data.load()
        np.save(path, weights)
        with path_lst.open("wb") as fp:
            pickle.dump(terms, fp)
        return terms, weights

    def tok2id(self, tok):
        return self._term2idx[tok]

    def id2tok(self, idx):
        return self._terms[idx]

    def lexicon_size(self) -> int:
        return len(self._terms)

    def dim(self):
        return self.embed.weight.shape[0]

    def forward(self, toks: TokenizedTexts):
        # lens ignored
        return self.embed(toks.ids + 1)  # +1 to handle padding at position -1


class WordvecUnkVocab(WordvecVocab):
    """Word-based embeddings with OOV

    A vocabulary in which all unknown terns are given the same token (UNK; 0)
    with random weights
    """

    def __post_init__(self):
        super().__post_init__()
        self._terms = [None] + self._terms
        for term in self._term2idx:
            self._term2idx[term] += 1
        unk_weights = self.random.state.normal(
            scale=0.5, size=(1, self._weights.shape[1])
        )
        self._weights = np.concatenate([unk_weights, self._weights])

    def tok2id(self, tok):
        return self._term2idx.get(tok, 0)

    def lexicon_path_segment(self):
        return "{base}_unk".format(base=super().lexicon_path_segment())

    def lexicon_size(self) -> int:
        return len(self._terms) + 1


class WordvecHashVocab(WordvecVocab):
    """Word-based embeddings with hash-based OOV

    A vocabulary in which all unknown terms are assigned a position in a
    flexible cache based on their hash value. Each position is assigned its own
    random weight.
    """

    hashspace: Param[int] = 1000
    init_stddev: Param[float] = 0.5
    log_miss: Param[bool] = False

    def __initialize__(self, options: ModuleInitOptions):
        super().__initialize__(options)

        size = (self.hashspace, self._weights.shape[1])
        if options.mode == ModuleInitMode.RANDOM:
            hash_weights = options.random.normal(scale=self.init_stddev, size=size)
        else:
            hash_weights: np.empty(size)

        self._weights = np.concatenate([self._weights, hash_weights])

    def tok2id(self, tok):
        try:
            return super().tok2id(tok)
        except KeyError:
            if self.log_miss:
                self.logger.debug(f"vocab miss {tok}")
            # NOTE: use md5 hash (or similar) here because hash() is not
            # consistent across runs
            item = tok.encode()
            item_hash = int(hashlib.md5(item).hexdigest(), 16)
            item_hash_pos = item_hash % self.hashspace
            return len(self._terms) + item_hash_pos

    def lexicon_size(self) -> int:
        return len(self._terms) + self.hashspace



# See documentation on https://datamaestro.readthedocs.io
from hashlib import md5
from datamaestro.definitions import dataset
from datamaestro.download.single import filedownloader
from datamaestro.utils import HashCheck
from xpmir.letor.distillation.samplers import PairwiseDistillationSamplesTSV


@filedownloader(
    "bert_cat_ensemble_msmarcopassage_train_scores_ids.tsv",
    "https://zenodo.org/record/4068216/files/"
    "bert_cat_ensemble_msmarcopassage_train_scores_ids.tsv?download=1",
    checker=HashCheck("4d99696386f96a7f1631076bcc53ac3c", md5),
)
@dataset(
    PairwiseDistillationSamplesTSV,
    url="https://github.com/sebastian-hofstaetter/neural-ranking-kd",
)
def msmarco_ensemble_teacher(bert_cat_ensemble_msmarcopassage_train_scores_ids):
    """Training files without the text content instead using the ids from MSMARCO

    The teacher files (using the data from "Train Triples Small" with ~40
    million triples) with the format pos_score neg_score query_id pos_passage_id
    neg_passage_id (with tab separation)
    """
    return {
        "with_docid": True,
        "with_queryid": True,
        "path": bert_cat_ensemble_msmarcopassage_train_scores_ids,
    }


@filedownloader(
    "bertbase_cat_msmarcopassage_train_scores_ids.tsv",
    "https://zenodo.org/record/4068216/files/"
    "bertbase_cat_msmarcopassage_train_scores_ids.tsv?download=1",
    checker=HashCheck("a2575af08a19b47c2041e67c9efcd917", md5),
)
@dataset(
    PairwiseDistillationSamplesTSV,
    url="https://github.com/sebastian-hofstaetter/neural-ranking-kd",
)
def msmarco_bert_teacher(bertbase_cat_msmarcopassage_train_scores_ids):
    """Training files without the text content instead using the ids from MSMARCO

    The teacher files (using the data from "Train Triples Small" with ~40
    million triples) with the format pos_score neg_score query_id pos_passage_id
    neg_passage_id (with tab separation)
    """
    return {
        "with_docid": True,
        "with_queryid": True,
        "path": bertbase_cat_msmarcopassage_train_scores_ids,
    }



from abc import ABC, abstractmethod
from typing import Generic, List, Tuple, TypeVar, Union, Optional, Callable
import sys

from attrs import define
from experimaestro import Param
import torch
import torch.nn as nn

from xpmir.learning.optim import Module

from xpmir.utils.utils import EasyLogger
from .tokenizers import (
    Tokenizer,
    TokenizedTexts,
    TokenizerBase,
    TokenizerOutput,
    TokenizerOptions,
)

T = TypeVar("T")


class Encoder(Module, EasyLogger, ABC):
    """Base class for all word and text encoders"""

    def __initialize__(self, options):
        # Easy and hacky way to get the device
        super().__initialize__(options)
        self._dummy_params = nn.Parameter(torch.Tensor())

    def static(self):
        return True

    @property
    def device(self):
        return self._dummy_params.device


@define
class TokensEncoderOutput:
    """Output representation for text encoder"""

    tokenized: TokenizedTexts
    """Tokenized texts"""

    value: torch.Tensor
    """The encoder output"""


class TokensEncoder(Tokenizer, Encoder):
    """(deprecated) Represent a text as a sequence of token representations"""

    def enc_query_doc(
        self, queries: List[str], documents: List[str], d_maxlen=None, q_maxlen=None
    ):
        """
        Returns encoded versions of the query and document from two
        list (same size) of queries and documents

        May be overwritten in subclass to provide contextualized representation, e.g.
        joinly modeling query and document representations in BERT.
        """

        tokenized_queries = self.batch_tokenize(queries, maxlen=q_maxlen)
        tokenized_documents = self.batch_tokenize(documents, maxlen=d_maxlen)
        return (
            tokenized_queries,
            self(tokenized_queries),
            tokenized_documents,
            self(tokenized_documents),
        )

    def forward(self, tokenized: TokenizedTexts):
        """
        Returns embeddings for the tokenized texts.

        tokenized: tokenized texts
        """
        raise NotImplementedError()

    def emb_views(self) -> int:
        """
        Returns how many "views" are returned by the embedding layer. Most have
        1, but sometimes it's useful to return multiple, e.g., BERT's multiple
        layers
        """
        return 1

    def dim(self) -> int:
        """
        Returns the number of dimensions of the embedding
        """
        raise NotImplementedError(f"for {self.__class__}")

    def static(self) -> bool:
        """
        Returns True if the representations are static, i.e., not trained.
        Otherwise False. This allows models to know when caching is appropriate.
        """
        return True

    def maxtokens(self) -> Optional[int]:
        """Maximum number of tokens that can be processed"""
        return None


LegacyEncoderInput = Union[List[str], List[Tuple[str, str]], List[Tuple[str, str, str]]]


InputType = TypeVar("InputType")
EncoderOutput = TypeVar("EncoderOutput")


class TextEncoderBase(Encoder, Generic[InputType, EncoderOutput]):
    """Base class for all text encoders"""

    __call__: Callable[Tuple["TextEncoderBase", List[InputType]], EncoderOutput]

    @abstractmethod
    def forward(self, texts: InputType) -> EncoderOutput:
        raise NotImplementedError()

    @property
    @abstractmethod
    def dimension(self) -> int:
        """Returns the dimension of the output space"""
        raise NotImplementedError()

    def max_tokens(self):
        """Returns the maximum number of tokens this encoder can process"""
        return sys.maxsize


class TextEncoder(TextEncoderBase[str, torch.Tensor]):
    """Encodes a text into a vector

    .. deprecated:: 1.3
        Use TextEncoderBase directly
    """

    pass


class DualTextEncoder(TextEncoderBase[Tuple[str, str], torch.Tensor]):
    """Encodes a pair of text into a vector

    .. deprecated:: 1.3
        Use TextEncoderBase directly
    """

    pass


class TripletTextEncoder(TextEncoderBase[Tuple[str, str, str], torch.Tensor]):
    """Encodes a triplet of text into a vector

    .. deprecated:: 1.3
        Use TextEncoderBase directly

    This is used in models such as DuoBERT where we encode (query, positive,
    negative) triplets.
    """

    pass


# --- Generic tokenized text encoders


@define
class RepresentationOutput:
    value: torch.Tensor
    """An arbitrary representation (by default, the batch dimension is the
    first)"""

    def __len__(self):
        return len(self.value)

    def __getitem__(self, ix: Union[slice, int]):
        return self.__class__(self.value[ix])

    @property
    def device(self):
        return self.value.device


@define
class TokensRepresentationOutput(RepresentationOutput):
    """A 3D tensor (batch x tokens x dimension)"""

    tokenized: TokenizedTexts
    """Tokenized texts"""

    def __getitem__(self, ix: Union[slice, int]):
        return self.__class__(self.value[ix], self.tokenized[ix])


@define
class TextsRepresentationOutput(RepresentationOutput):
    """Value is atensor representing full texts (batch x dimension)"""

    tokenized: TokenizedTexts
    """Tokenized texts"""

    def to(self, device):
        return self.__class__(self.value.to(device), self.tokenized.to(device))

    def __getitem__(self, ix: Union[slice, int]):
        return self.__class__(self.value[ix], self.tokenized[ix])


class TokenizedEncoder(Encoder, Generic[EncoderOutput, TokenizerOutput]):
    """Encodes a tokenized text into a vector"""

    @abstractmethod
    def forward(self, inputs: TokenizerOutput) -> EncoderOutput:
        pass

    @property
    def max_length(self):
        """Returns the maximum length that the model can process"""
        return sys.maxsize


class TokenizedTextEncoderBase(TextEncoderBase[InputType, EncoderOutput]):
    @abstractmethod
    def forward(
        self, inputs: List[InputType], options: Optional[TokenizerOptions] = None
    ) -> EncoderOutput:
        ...


class TokenizedTextEncoder(
    TokenizedTextEncoderBase[InputType, EncoderOutput],
    Generic[InputType, EncoderOutput, TokenizerOutput],
):
    """Encodes a tokenizer input into a vector

    This pipelines two objects:

    1. A tokenizer that segments the text;
    2. An encoder that returns a representation of the tokens in a vector space
    """

    tokenizer: Param[TokenizerBase[InputType, TokenizerOutput]]
    encoder: Param[TokenizedEncoder[TokenizerOutput, EncoderOutput]]

    def __initialize__(self, options):
        super().__initialize__(options)
        self.tokenizer.initialize(options)
        self.encoder.initialize(options)

    def forward(
        self, inputs: List[InputType], *args, options: Optional[TokenizerOptions] = None
    ) -> EncoderOutput:
        assert len(args) == 0, "Unhandled extra arguments"
        tokenized = self.tokenizer.tokenize(inputs, options)
        return self.forward_tokenized(tokenized, *args)

    def forward_tokenized(self, tokenized):
        return self.encoder(tokenized)

    def tokenize(
        self, inputs: List[InputType], options: Optional[TokenizerOptions] = None
    ):
        options = options or TokenizerOptions()
        options.max_length = min(self.encoder.max_length, options.max_length or None)
        return self.tokenizer.tokenize(inputs, options)

    def static(self):
        """Whether embeddings parameters are learnable"""
        return self.encoder.static()

    @property
    def dimension(self):
        return self.encoder.dimension



from abc import abstractmethod
from typing import Iterable, Optional, List
import torch
from experimaestro import Param, Annotated, field
from xpmir.learning.context import TrainerContext
from xpmir.neural.dual import (
    DualVectorScorer,
    TopicRecord,
    DocumentRecord,
)
from xpmir.text import TokenizedTextEncoderBase, TokenizerOptions, TokensEncoderOutput
from xpmir.text.encoders import InputType
from .common import SimilarityInput, SimilarityOutput, Similarity


class InteractionScorer(DualVectorScorer[SimilarityInput, SimilarityInput]):
    """Interaction-based neural scorer

    This is the base class for all scorers that depend on a map
    of cosine/inner products between query and document token representations.
    """

    encoder: Param[TokenizedTextEncoderBase[InputType, TokensEncoderOutput]]
    """The embedding model -- the vocab also defines how to tokenize text"""

    query_encoder: Annotated[
        Optional[TokenizedTextEncoderBase[InputType, TokensEncoderOutput]],
        field(default=None),
    ]
    """The embedding model for queries (if None, uses encoder)"""

    similarity: Param[Similarity]
    """Which similarity function to use - ColBERT uses a cosine similarity by default"""

    qlen: Param[int] = 20
    """Maximum query length (this can be even shortened by the model)"""

    dlen: Param[int] = 2000
    """Maximum document length (this can be even shortened by the model)"""

    def __validate__(self):
        super().__validate__()
        assert (
            self.dlen <= self.encoder.max_tokens()
        ), f"The maximum document length ({self.dlen}) should be less "
        "that what the vocab can process ({self.encoder.max_tokens()})"
        assert (
            self.qlen <= self.encoder.max_tokens()
        ), f"The maximum query length ({self.qlen}) should be less "
        "that what the vocab can process ({self.encoder.max_tokens()})"

    def _encode(
        self,
        texts: List[InputType],
        encoder: TokenizedTextEncoderBase[InputType, TokensEncoderOutput],
        options: TokenizerOptions,
    ) -> SimilarityInput:
        encoded = encoder(texts, options=options)
        return self.similarity.preprocess(
            SimilarityInput(encoded.value, encoded.tokenized.mask)
        )

    def encode_documents(self, records: Iterable[DocumentRecord]) -> SimilarityInput:
        return self.similarity.preprocess(
            self._encode(
                records,
                self.encoder,
                TokenizerOptions(self.dlen),
            )
        )

    def encode_queries(self, records: Iterable[TopicRecord]) -> SimilarityInput:
        return self.similarity.preprocess(
            self._encode(
                records,
                self._query_encoder,
                TokenizerOptions(self.qlen),
            )
        )

    def score_pairs(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
        info: Optional[TrainerContext] = None,
    ) -> torch.Tensor:
        similarity = self.similarity.compute_pairs(queries, documents)
        return self.compute_scores(queries, documents, similarity, info)

    def score_product(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
        info: Optional[TrainerContext] = None,
    ) -> torch.Tensor:
        similarity = self.similarity.compute_product(queries, documents)
        return self.compute_scores(queries, documents, similarity, info)

    @abstractmethod
    def compute_scores(
        self,
        queries: SimilarityInput,
        documents: SimilarityInput,
        similarity: SimilarityOutput,
        info: Optional[TrainerContext] = None,
    ):
        ...



from abc import ABC, abstractmethod
from typing import List
from datamaestro.record import Record
from datamaestro_text.data.conversation import (
    DecontextualizedItem,
)
from xpmir.text.encoders import TextEncoderBase
from xpmir.letor.trainers.alignment import RepresentationOutput
from xpmir.utils.convert import Converter


class ConversationRepresentationEncoder(
    TextEncoderBase[List[Record], RepresentationOutput], ABC
):
    @abstractmethod
    def forward(self, conversations: List[Record]) -> RepresentationOutput:
        """Represents a list of conversations"""
        ...


class DecontextualizedQueryConverter(Converter[Record, str]):
    def __call__(self, input: Record) -> str:
        return input[DecontextualizedItem].get_decontextualized_query()



from typing import Optional
from xpmir.text.huggingface import TransformerTokensEncoder
from xpmir.neural.splade import SpladeTextEncoder, MaxAggregation, DotDense
from experimaestro.huggingface import ExperimaestroHFHub


def push_splade(
    variant: str, model_id: str, *, query_model_id: Optional[str] = None, config=None
):
    encoder = TransformerTokensEncoder(model_id=model_id, trainable=True)
    query_encoder = (
        TransformerTokensEncoder(model_id=query_model_id, trainable=True)
        if query_model_id
        else None
    )
    aggregation = MaxAggregation()

    # make use the output of the BERT and do an aggregation
    doc_encoder = SpladeTextEncoder(
        aggregation=aggregation, encoder=encoder, maxlen=200
    )
    query_encoder = SpladeTextEncoder(
        aggregation=aggregation, encoder=encoder, maxlen=31
    )

    model = DotDense(encoder=doc_encoder, query_encoder=query_encoder)

    ExperimaestroHFHub(model, variant=variant).push_to_hub(
        repo_id="xpmir/splade", config=config
    )


config = {
    "variants": [
        "cocondenser-selfdistil",
        "cocondenser-ensembledistil",
        "efficient-V-large-doc",
        "efficient-VI-BT-large-doc",
    ]
}

push_splade(
    "cocondenser-selfdistil", "naver/splade-cocondenser-selfdistil", config=config
)
push_splade("cocondenser-ensembledistil", "naver/splade-cocondenser-ensembledistil")
push_splade(
    "efficient-V-large-doc",
    "naver/efficient-splade-V-large-doc",
    query_model_id="naver/efficient-splade-V-large-query",
)
push_splade(
    "efficient-VI-BT-large-doc",
    "naver/efficient-splade-VI-BT-large-doc",
    query_model_id="naver/efficient-splade-VI-BT-large-query",
)



from pathlib import Path
from typing import Iterator, Optional
import torch
import torch.nn as nn
from experimaestro import Param
from xpmir.learning.optim import Module, ModuleLoader
from xpmir.learning.parameters import (
    ParameterElement,
    ParametersIterator,
    SubParametersIterator,
    InverseParametersIterator,
    SubModuleLoader,
)


class Model(Module, nn.Module):
    sub_model: Param[Optional["Model"]]

    def __post_init__(self):
        self.layer = nn.Linear(1, 1)


class MyParametersIterator(ParametersIterator):
    model: Param[Model]

    def iter(self) -> Iterator[ParameterElement]:
        for name, param in self.model.named_parameters():
            yield ParameterElement(
                name, self.model, param, param is self.model.layer.weight
            )


def test_iterator():
    model = Model(sub_model=Model())
    pi = MyParametersIterator(model=model)

    for x in pi.instance().iter():
        assert x.selected == (x.name == "layer.weight")


def test_inverse_iterator():
    model = Model(sub_model=Model())
    pi = InverseParametersIterator(iterator=MyParametersIterator(model=model))

    for x in pi.instance().iter():
        assert x.selected == (x.name != "layer.weight")


def test_sub_iterator():
    sub = Model()
    model = Model(sub_model=sub)
    sub_iterator = MyParametersIterator(model=sub)
    pi = SubParametersIterator(
        model=model, default=False, iterator=sub_iterator
    ).instance()

    for x in pi.iter():
        assert x.selected == (x.name == "sub_model.layer.weight")


def test_SubModuleLoader(tmp_path: Path):
    """Test the partial module loader"""
    sub = Model()
    model = Model(sub_model=sub)
    datapath = tmp_path / "data.pth"
    loader = ModuleLoader(path=datapath, value=model)
    loader.__xpm__.task = loader  # For testing

    # Use an instance and save the parameters
    model_instance = model.instance()
    with datapath.open("wb") as fp:
        torch.save(model_instance.state_dict(), fp)

    # creates an instance with loader
    sub = Model()
    model = Model(sub_model=sub)
    sub_instance = sub.instance()
    assert model_instance.sub_model.layer.weight != sub_instance.layer.weight

    # creates an instance without loader
    sub = Model()
    model = Model(sub_model=sub)
    sub_selector = MyParametersIterator(model=sub)
    sub_loader = SubModuleLoader.from_module_loader(loader, model, sub, sub_selector)
    sub.add_pretasks(sub_loader)
    sub_instance = sub.instance()
    assert model_instance.sub_model.layer.weight == sub_instance.layer.weight



from dataclasses import dataclass
from enum import Enum
import threading
from typing import Any, Callable, List, Optional, TYPE_CHECKING, Union
from pathlib import Path
import numpy as np
import torch
import re

from experimaestro import (
    Config,
    Param,
    tagspath,
    Task,
    PathSerializationLWTask,
    experiment,
    RunMode,
)
from experimaestro.scheduler import Job, Listener
from experimaestro.utils import cleanupdir
from xpmir.utils.logging import easylog, LazyJoin
from experimaestro.scheduler.services import WebService
from xpmir.context import Hook, Context
from xpmir.utils.utils import Initializable, foreach
from xpmir.learning.metrics import ScalarMetric
from .schedulers import Scheduler

if TYPE_CHECKING:
    from xpmir.learning.context import TrainerContext

logger = easylog()


class Optimizer(Config):
    def __call__(self, parameters) -> torch.optim.Optimizer:
        raise NotImplementedError()


class SGD(Optimizer):
    """Wrapper for SGD optimizer in Pytorch"""

    lr: Param[float] = 1e-5
    """Learning rate"""

    weight_decay: Param[float] = 0.0
    """Weight decay (L2)"""

    def __call__(self, parameters):
        from torch.optim import SGD

        return SGD(parameters, lr=self.lr, weight_decay=self.weight_decay)


class Adafactor(Optimizer):
    """Wrapper for Adafactor optimizer in Transformers library

    See :class:`transformers.optimization.Adafactor` for full documentation
    """

    lr: Param[Optional[float]] = None
    """Learning rate"""

    weight_decay: Param[float] = 0.0
    """Weight decay (L2)"""

    relative_step: Param[bool] = True
    """If true, time-dependent learning rate is computed instead of external
    learning rate"""

    def __call__(self, parameters):
        from transformers.optimization import Adafactor

        return Adafactor(
            parameters,
            lr=self.lr,
            weight_decay=self.weight_decay,
            relative_step=self.relative_step,
        )


class Adam(Optimizer):
    """Wrapper for Adam optimizer in PyTorch"""

    lr: Param[float] = 1e-3
    """Learning rate"""

    weight_decay: Param[float] = 0.0
    """Weight decay (L2)"""

    eps: Param[float] = 1e-8

    def __call__(self, parameters):
        from torch.optim import Adam

        return Adam(
            parameters, lr=self.lr, weight_decay=self.weight_decay, eps=self.eps
        )


class AdamW(Optimizer):
    """Adam optimizer that takes into account the regularization

    See the `PyTorch documentation
    <https://pytorch.org/docs/stable/generated/torch.optim.AdamW.html>`_
    """

    lr: Param[float] = 1e-3
    weight_decay: Param[float] = 1e-2
    eps: Param[float] = 1e-8

    def __call__(self, parameters):
        from torch.optim import AdamW

        return AdamW(
            parameters, lr=self.lr, weight_decay=self.weight_decay, eps=self.eps
        )


class ModuleInitMode(Enum):
    """Initialization mode"""

    #: Default initialization (i.e. can load default parameters or initialize randomly)
    DEFAULT = 0

    #: No parameter initialization (just initialize the structure of the model)
    NONE = 1

    #: Random initialization (initialize the structure, then use a the random
    #: number generator to initialize the values)
    RANDOM = 2

    def to_options(self, random: Optional[np.random.RandomState] = None):
        return ModuleInitOptions(self, random)


@dataclass
class ModuleInitOptions:
    #: Initialization mode
    mode: ModuleInitMode

    #: Random generator (only defined when mode is RANDOM)
    random: Optional[np.random.RandomState] = None


class Module(Config, Initializable, torch.nn.Module):
    """A module contains parameters"""

    def __init__(self):
        Initializable.__init__(self)
        torch.nn.Module.__init__(self)

    def __initialize__(self, options: ModuleInitOptions):
        """Initialize a module

        :param options: The initialization options
        """
        pass

    def __call__(self, *args, **kwargs):
        return torch.nn.Module.__call__(self, *args, **kwargs)

    def to(self, *args, **kwargs):
        return torch.nn.Module.to(self, *args, **kwargs)


class ModuleList(Module, Initializable):
    """Groups different models together, to be used within the Learner"""

    sub_modules: Param[List[Module]]

    def __post_init__(self):
        # Register sub-modules
        for ix, sub_module in enumerate(self.sub_modules):
            self.add_module(str(ix), sub_module)

    def __initialize__(self, options: ModuleInitOptions):
        for module in self.sub_modules:
            module.initialize(options)

    def __call__(self, *args, **kwargs):
        raise AssertionError("This module cannot be used as such")

    def to(self, *args, **kwargs):
        return torch.nn.Module.to(self, *args, **kwargs)


class ModuleLoader(PathSerializationLWTask):
    def execute(self):
        """Loads the model from disk using the given serialization path"""
        logger.info("Loading model from disk: %s", self.path)
        self.value.initialize(ModuleInitMode.NONE.to_options())
        data = torch.load(self.path)
        self.value.load_state_dict(data)


class ParameterFilter(Config):
    """One abstract class which doesn't do the filtrage"""

    def __call__(self, name, params) -> bool:
        """Returns true if the parameters should be optimized with the
        associated optimizer"""
        return True


class RegexParameterFilter(ParameterFilter):
    """gives the name of the model to do the filtrage
    Precondition: Only and just one of the includes and excludes can be None"""

    includes: Param[Optional[List[str]]] = None
    """The str of params to be included from the model"""

    excludes: Param[Optional[List[str]]] = None
    """The str of params to be excludes from the model"""

    def __init__(self):
        self.name = set()

    def __validate__(self):
        return self.includes or self.excludes

    def __repr__(self) -> str:
        return f"RegexParameterFilter({self.includes}, {self.excludes})"

    def __call__(self, name, params) -> bool:
        # Look first at included
        if self.includes:
            for regex in self.includes:
                if re.search(regex, name):
                    return True

            # Include if not excluded
            if not self.excludes:
                return False

        for regex in self.excludes:
            if re.search(regex, name):
                return False

        return True


class ParameterOptimizer(Config):
    """Associates an optimizer with a list of parameters to optimize"""

    optimizer: Param[Optimizer]
    """The optimizer"""

    scheduler: Param[Optional[Scheduler]]
    """The optional scheduler"""

    module: Param[Optional[Module]]
    """The module from which parameters should be extracted"""

    filter: Param[Optional[ParameterFilter]] = ParameterFilter.C()
    """How parameters should be selected for this (by default, use them all)"""

    def create_optimizer(
        self, module: Module, filter: Callable[[str, Any], bool]
    ) -> torch.optim.Optimizer:
        """Returns a (pytorch) optimizer"""
        module = self.module or module
        params = {
            name: param
            for name, param in module.named_parameters()
            if (self.filter is None or self.filter(name, param)) and filter(name, param)
        }
        if not params:
            logger.warning(
                "Parameter list: %s", [name for name, _ in module.named_parameters()]
            )
            raise RuntimeError(f"Parameter list is empty with {self.filter}")

        logger.debug(
            "Optimizing with %s parameters [%s]",
            self.filter,
            LazyJoin(",", params.keys()),
        )
        optimizer = self.optimizer(params.values())
        return optimizer


class DuplicateParameterFilter:
    """Filters out already optimized parameters"""

    def __init__(self):
        self.parameters = set()

    def __call__(self, name, params):
        if params in self.parameters:
            return False
        self.parameters.add(params)
        return True


class OptimizationHook(Hook):
    """Base class for all optimization hooks"""

    pass


class GradientHook(OptimizationHook):
    """Hooks that are called when the gradient is computed

    The gradient is guaranteed to be unscaled in this case.
    """

    pass


class GradientClippingHook(GradientHook):
    """Gradient clipping"""

    max_norm: Param[float]
    """Maximum norm for gradient clipping"""

    def __call__(self, main: "ScheduledOptimizer"):
        torch.nn.utils.clip_grad_norm_(main.module.parameters(), self.max_norm)


class GradientLogHook(GradientHook):
    """ "Log the gradient norm"""

    name: Param[str] = "gradient_norm"

    def __call__(self, main: "ScheduledOptimizer"):
        sum_norms = 0.0
        n_params = 0
        with torch.no_grad():
            for param in main.module.parameters():
                if param.grad is not None:
                    n_params += param.grad.numel()
                    sum_norms += param.grad.numel() * param.grad.norm() ** 2

        main.trainer_context.writer.add_scalar(
            self.name, sum_norms / n_params, main.trainer_context.state.step
        )


class ScheduledOptimizer:
    def initialize(
        self,
        param_optimizers: List[ParameterOptimizer],
        num_training_steps: int,
        module: Module,
        use_scaler: bool,
        hooks: List[OptimizationHook] = [],
        trainer_context: Optional["TrainerContext"] = None,
    ):
        self.schedulers = []
        self.scheduler_factories = []
        self.optimizers = []
        self.scheduler_steps = -1  # Number of scheduler steps
        self.num_training_steps = num_training_steps
        self.module = module
        self.context = Context(hooks)
        self.trainer_context = trainer_context

        try:
            next(module.parameters())
        except StopIteration:
            raise RuntimeError(f"No parameters to optimize in the module {module}")

        filter = DuplicateParameterFilter()
        for param_optimizer in param_optimizers:
            optimizer = param_optimizer.create_optimizer(module, filter)
            self.optimizers.append(optimizer)
            self.scheduler_factories.append(param_optimizer.scheduler)

        self.reset_schedulers()

        assert len(self.schedulers) == len(self.optimizers)

        if use_scaler:
            logger.info("Using GradScaler when optimizing")
        self.scaler = torch.amp.GradScaler('cuda') if use_scaler else None

    def load_state_dict(self, state):
        for optimizer, optimizer_state in zip(self.optimizers, state["optimizers"]):
            optimizer.load_state_dict(optimizer_state)

        if self.scaler is not None:
            self.scaler.load_state_dict(state["scaler"])

        # Re-create schedulers
        self.scheduler_steps = state["scheduler_steps"]
        self.reset_schedulers()

    def reset_schedulers(self):
        self.schedulers = []
        for optimizer, scheduler_factory in zip(
            self.optimizers, self.scheduler_factories
        ):
            if scheduler_factory is None:
                self.schedulers.append(None)
            else:
                self.schedulers.append(
                    scheduler_factory(
                        optimizer,
                        self.num_training_steps,
                        last_epoch=self.scheduler_steps,
                    )
                )

    def state_dict(self):
        return {
            "optimizers": [optimizer.state_dict() for optimizer in self.optimizers],
            "scaler": None if self.scaler is None else self.scaler.state_dict(),
            "scheduler_steps": self.scheduler_steps,
        }

    def scale(self, loss: torch.Tensor):
        if self.scaler is None:
            return loss
        return self.scaler.scale(loss)

    def zero_grad(self):
        """Zero-grad for all optimizers"""
        for optimizer in self.optimizers:
            optimizer.zero_grad()

    def optimizer_step(self, context: "TrainerContext"):
        """Performs an optimizer step (using the scaler if defined)"""
        if self.scaler is None:
            # Apply gradient hooks
            foreach(
                self.context.hooks(GradientHook),
                lambda hook: hook(self),
            )

            for optimizer in self.optimizers:
                optimizer.step()

        else:
            # Unscale first
            for optimizer in self.optimizers:
                self.scaler.unscale_(optimizer)

            # Apply gradient hooks
            foreach(
                self.context.hooks(GradientHook),
                lambda hook: hook(self),
            )

            # Step
            for optimizer in self.optimizers:
                self.scaler.step(optimizer)
            context.add_metric(
                ScalarMetric("gradient/scaler", self.scaler.get_scale(), 1)
            )
            self.scaler.update()

    def scheduler_step(self, context: "TrainerContext"):
        """Performs a step for all the schedulers"""
        for ix, scheduler in enumerate(self.schedulers):
            if scheduler is not None:
                for p_ix, lr in enumerate(scheduler.get_last_lr()):
                    context.add_metric(
                        ScalarMetric(f"gradient/scheduler/{ix+1}/{p_ix+1}", lr, 1)
                    )
                scheduler.step()
        self.scheduler_steps += 1


Optimizers = Union[ParameterOptimizer, Optimizer, List[ParameterOptimizer]]
"""Defines a set of optimizers"""


def get_optimizers(optimizers: Optimizers):
    """Returns a list of ParameterOptimizer"""
    if isinstance(optimizers, list):
        return optimizers

    if isinstance(optimizers, ParameterOptimizer):
        return [optimizers]

    return [ParameterOptimizer(optimizer=optimizers)]


class TensorboardServiceListener(Listener):
    def __init__(self, source: Path, target: Path):
        self.source = source
        self.target = target

    def job_state(self, job: Job):
        if not job.state.notstarted():
            if not self.source.is_symlink():
                try:
                    self.source.symlink_to(self.target)
                except Exception:
                    logger.exception(
                        "Cannot symlink %s to %s", self.source, self.target
                    )


class TensorboardService(WebService):
    id = "tensorboard"

    def __init__(self, xp: experiment, path: Path):
        super().__init__()

        self.path = path
        self.url = None
        self.run_mode = xp.run_mode

        if self.run_mode == RunMode.NORMAL:
            cleanupdir(self.path)
            self.path.mkdir(exist_ok=True, parents=True)
            logger.info("You can monitor learning with:")
            logger.info("tensorboard --logdir=%s", self.path)

    def add(self, task: Task, path: Path):
        # Wait until config has started
        if self.run_mode == RunMode.NORMAL:
            if job := task.__xpm__.job:
                if job.scheduler is not None:
                    tag_path = tagspath(task)
                    if tag_path:
                        job.scheduler.addlistener(
                            TensorboardServiceListener(self.path / tag_path, path)
                        )
                    else:
                        logger.error(
                            "The task is not associated with tags: "
                            "cannot link to tensorboard data"
                        )
                else:
                    logger.debug("No scheduler: not adding the tensorboard data")
            else:
                logger.error(
                    "Task was not started: cannot link to tensorboard job path"
                )

    def description(self):
        return "Tensorboard service"

    def close(self):
        if self.server and self.run_mode == RunMode.NORMAL:
            self.server.shutdown()

    def _serve(self, running: threading.Event):
        if self.run_mode != RunMode.NORMAL:
            return

        import tensorboard as tb

        try:
            logger.info("Starting %s service", self.id)
            self.program = tb.program.TensorBoard()
            self.program.configure(
                host="localhost",
                logdir=str(self.path.absolute()),
                path_prefix=f"/services/{self.id}",
                port=0,
            )
            self.server = self.program._make_server()

            self.url = self.server.get_url()
            running.set()
            self.server.serve_forever()
        except Exception:
            logger.exception("Error while starting tensorboard")
            running.set()



from typing import (
    Generic,
    List,
    Protocol,
    Callable,
    Iterator,
    TypeVar,
    Union,
    Sequence,
)
from experimaestro import Config
from pytorch_lightning.utilities.memory import garbage_collection_cuda
from xpmir.utils.utils import easylog

logger = easylog()

RT = TypeVar("RT")
T = TypeVar("T")
ARGS = TypeVar("ARGS")
KWARGS = TypeVar("KWARGS")


class RecoverableOOMError(Exception):
    """Exception raised when a OOM occurs and the batcher
    is taking this into account for the next round (i.e. we
    can reprocess the batch with a higher probability of not
    having an OOM)
    """

    pass


class IterativeProcessor(Protocol, Generic[T, RT, ARGS, KWARGS]):
    def __call__(
        self, batch: Iterator[Sequence[T]], length: int, *args: ARGS, **kwargs: KWARGS
    ) -> RT:
        """Process a series of batches

        :argument length: The number of batches
        """
        ...


class Processor(Protocol, Generic[T, ARGS, KWARGS]):
    def __call__(self, batch: Sequence[T], *args: ARGS, **kwargs: KWARGS) -> None:
        ...


class Reducer(Protocol, Generic[T, RT, ARGS, KWARGS]):
    def __call__(
        self, batch: Sequence[T], value: RT, *args: ARGS, **kwargs: KWARGS
    ) -> RT:
        ...


class BatcherWorker:
    def __init__(self, batch_size: int):
        self.batch_size = batch_size

    def process_withreplay(
        self,
        batch: Sequence[T],
        process: IterativeProcessor[T, RT, ARGS, KWARGS],
        *args: ARGS,
        **kwargs: KWARGS,
    ) -> RT:
        """Process a batch with replay

        Replay = if an error occurs, the full batch is re-processed
        """
        return process(iter([batch]), 1, *args, **kwargs)

    def process(
        self,
        batch: Sequence[T],
        process: Processor[T, ARGS, KWARGS],
        *args: ARGS,
        raise_oom=False,
        **kwargs: KWARGS,
    ) -> None:
        """Process a batch

        If a recoverable OOM error occurs, the processing continues but do not
        reprocess previously processed samples.

        Arguments:

            raise_oom: Raise an OOM exception when an OOM is recoverable
        """
        process(batch, *args, **kwargs)

    def reduce(
        self,
        batch: Sequence[T],
        reducer: Reducer[T, RT, ARGS, KWARGS],
        initialvalue: RT,
        *args: ARGS,
        raise_oom=False,
        **kwargs: KWARGS,
    ) -> RT:
        """Attributes:

        Arguments:
            batch: The data to process

            reducer: The reducer function, whose two first arguments are a slice
            of T and the reduced value, and that returns a new value

            raise_oom: Raise an OOM exception when an OOM is recoverable instead
            of continuing
        """
        return reducer(batch, initialvalue, *args, **kwargs)

    async def aio_reduce(
        self,
        batch: Sequence[T],
        reducer: Reducer[T, RT, ARGS, KWARGS],
        initialvalue: RT,
        *args: ARGS,
        raise_oom=False,
        **kwargs: KWARGS,
    ) -> RT:
        """Attributes:

        Arguments:
            batch: The data to process

            reducer: The reducer function, whose two first arguments are a slice
            of T and the reduced value, and that returns a new value

            raise_oom: Raise an OOM exception when an OOM is recoverable instead
            of continuing
        """
        return await reducer(batch, initialvalue, *args, **kwargs)


def is_cublas_alloc_failed(exception):
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "CUBLAS_STATUS_ALLOC_FAILED" in exception.args[0]
    )


def is_oom_error(exception: Exception):
    """Detect CUDA out-of-memory errors"""
    import torch
    from pytorch_lightning.utilities.memory import is_oom_error as legacy_check

    if isinstance(exception, torch.cuda.OutOfMemoryError):
        return True

    if legacy_check(exception):
        return True

    return is_cublas_alloc_failed(exception)


class Batcher(Config):
    """Responsible for micro-batching when the batch does not fit in memory

    The base class just does nothing (no adaptation)
    """

    def initialize(self, batch_size: int) -> BatcherWorker:
        logger.info("Using a simple batcher with batch size %d", batch_size)
        return BatcherWorker(batch_size)


class PowerAdaptativeBatcherWorker(BatcherWorker):
    def __init__(self, batch_size: int):
        super().__init__(batch_size)
        self.max_batch_size = batch_size
        self.current_divider = 1
        logger.info("Adaptative batcher: initial batch size is %d", self.batch_size)

    def get_ranges(self, batch_size):
        ranges = []
        ix = 0
        while ix < batch_size:
            ranges.append(slice(ix, ix + self.batch_size))
            ix += self.batch_size
        return ranges

    def iter(self, batch: Sequence[T], ranges: List[slice]) -> Iterator[Sequence[T]]:
        for range in ranges:
            yield batch[range]

    def process_withreplay(
        self,
        batch: Sequence[T],
        process: IterativeProcessor[T, RT, ARGS, KWARGS],
        *args: ARGS,
        **kwargs: KWARGS,
    ) -> RT:
        while True:
            ranges = self.get_ranges(len(batch))
            rt = self._run(
                lambda: process(self.iter(batch, ranges), len(ranges), *args, **kwargs)
            )
            if not isinstance(rt, RecoverableOOMError):
                return rt

    def process(
        self,
        batch: Sequence[T],
        process: Processor[T, ARGS, KWARGS],
        *args: ARGS,
        raise_oom=False,
        **kwargs: KWARGS,
    ) -> None:
        ix = 0
        while ix < len(batch):
            s = slice(ix, ix + self.batch_size)
            rt = self._run(lambda: process(batch[s], *args, **kwargs))
            if not isinstance(rt, RecoverableOOMError):
                ix += self.batch_size
            elif raise_oom:
                raise rt

    def reduce(
        self,
        batch: Sequence[T],
        reducer: Reducer[T, RT, ARGS, KWARGS],
        value: RT,
        *args: ARGS,
        raise_oom=False,
        **kwargs: KWARGS,
    ) -> RT:
        """
        Reduce a batch using the process function

        Args:
            batch: A batch
            reducer: A function that

        raise_oom: Raise an OOM exception when an OOM is recoverable
        """
        ix = 0
        while ix < len(batch):
            s = slice(ix, ix + self.batch_size)
            rt = self._run(lambda: reducer(batch[s], value, *args, **kwargs))
            if not isinstance(rt, RecoverableOOMError):
                ix += self.batch_size
                value = rt
            elif raise_oom:
                raise RecoverableOOMError()

        return value

    def _run(self, process: Callable[[], RT]) -> Union[RecoverableOOMError, RT]:
        try:
            # Perform a process
            return process()
        except RuntimeError as exception:
            if is_oom_error(exception):
                garbage_collection_cuda()
                while True:
                    self.current_divider += 1
                    new_batchsize = self.max_batch_size // self.current_divider
                    if new_batchsize != self.batch_size:
                        break

                self.batch_size = new_batchsize
                if self.batch_size == 0:
                    logger.error("Cannot decrease batch size below 1")
                    raise
                logger.info(
                    "Adaptative batcher: reducing batch size to %d", self.batch_size
                )
                return RecoverableOOMError(f"Reducing batch size to {self.batch_size}")
            else:
                # Other exception
                raise


class PowerAdaptativeBatcher(Batcher):
    """Starts with the provided batch size, and then divides in 2, 3, etc.
    until there is no more OOM
    """

    def initialize(self, batch_size: int) -> PowerAdaptativeBatcherWorker:
        return PowerAdaptativeBatcherWorker(batch_size)



from datamaestro.record import record_type
import pytest
import numpy as np
from typing import Iterator, Tuple
from experimaestro import Param
import datamaestro_text.data.ir as ir
from xpmir.rankers import Retriever
from xpmir.letor.samplers import (
    TrainingTriplets,
    TripletBasedSampler,
    ModelBasedSampler,
)
from xpmir.documents.samplers import RandomSpanSampler

# ---- Serialization


class MyTrainingTriplets(TrainingTriplets):
    def iter(
        self,
    ) -> Iterator[Tuple[ir.TopicRecord, ir.DocumentRecord, ir.DocumentRecord]]:
        count = 0

        while True:
            yield ir.create_record(text=f"q{count}"), ir.create_record(
                id=1, text=f"doc+{count}"
            ), ir.create_record(id=2, text=f"doc-{count}")

    topic_recordtype = record_type(ir.IDItem, ir.SimpleTextItem)
    document_recordtype = record_type(ir.SimpleTextItem)


def test_serializing_tripletbasedsampler():
    """Serialized samplers should start back from the saved state"""
    # Collect samples and state after 10 samples
    sampler = TripletBasedSampler(
        source=MyTrainingTriplets(id="test-triplets")
    ).instance()
    iter = sampler.pairwise_iter()

    for _, _ in zip(range(10), iter):
        pass
    data = iter.state_dict()

    samples = []
    for _, record in zip(range(10), sampler.pairwise_iter()):
        samples.append(record)

    # Test
    sampler = TripletBasedSampler(
        source=MyTrainingTriplets(id="test-triplets")
    ).instance()
    iter = sampler.pairwise_iter()
    iter.load_state_dict(data)
    for _, record, expected in zip(range(10), iter, samples):
        assert expected.query[ir.TextItem].text == record.query[ir.TextItem].text
        assert expected.positive[ir.TextItem].text == record.positive[ir.TextItem].text
        assert expected.negative[ir.TextItem].text == record.negative[ir.TextItem].text


class GeneratedDocuments(ir.Documents):
    pass


class GeneratedTopics(ir.Topics):
    pass


class GeneratedAssessments(ir.AdhocAssessments):
    pass


class RandomRetriever(Retriever):
    pass


def adhoc_synthetic_dataset():
    """Creates a random dataset"""
    return ir.Adhoc(
        documents=GeneratedDocuments(),
        topics=GeneratedTopics(),
        assessments=GeneratedAssessments(),
    )


@pytest.mark.skip("not yet done")
def test_modelbasedsampler():
    dataset = adhoc_synthetic_dataset()
    sampler = ModelBasedSampler(
        dataset=dataset, retriever=RandomRetriever(dataset=dataset)
    ).instance()

    for a in sampler._itertopics():
        pass


class FakeDocumentStore(ir.DocumentStore):
    id: Param[str] = ""

    @property
    def documentcount(self):
        return 10

    def document_int(self, internal_docid: int) -> ir.DocumentRecord:
        return ir.create_record(id=str(internal_docid), text=f"D{internal_docid} " * 10)


def test_pairwise_randomspansampler():
    documents = FakeDocumentStore()

    sampler1 = RandomSpanSampler(documents=documents).instance()

    sampler2 = RandomSpanSampler(documents=documents).instance()

    random1 = np.random.RandomState(seed=0)
    random2 = np.random.RandomState(seed=0)
    sampler1.initialize(random1)
    sampler2.initialize(random2)
    iter1 = sampler1.pairwise_iter()
    iter2 = sampler2.pairwise_iter()

    for s1, s2, _ in zip(iter1, iter2, range(10)):
        # check that they are the same with same random state
        assert s1.query[ir.TextItem].text == s2.query[ir.TextItem].text
        assert s1.positive[ir.TextItem].text == s2.positive[ir.TextItem].text
        assert s1.negative[ir.TextItem].text == s2.negative[ir.TextItem].text



from pathlib import Path
from experimaestro.tools.documentation import DocumentationAnalyzer


def test_documented():
    """Test if every configuration is documented"""
    doc_path = Path(__file__).parents[3] / "docs" / "source" / "index.rst"
    analyzer = DocumentationAnalyzer(doc_path, set(["xpmir"]), set(["xpmir.test"]))

    analyzer.analyze()
    analyzer.report()
    analyzer.assert_valid_documentation()



from pathlib import Path
from typing import Dict, Iterator
import torch
from experimaestro import DirectoryContext
from xpmir.learning import Random
from xpmir.learning.learner import Learner, Module
from xpmir.learning.optim import Adam, get_optimizers
from xpmir.learning.trainers import Trainer
from experimaestro.taskglobals import Env as TaskEnv


class TheModel(Module):
    def __post_init__(self):
        self.layer = torch.nn.Linear(3, 5)

    def __initialize__(self, *args):
        pass


class TheTrainer(Trainer):
    def __post_init__(self):
        super().__post_init__()
        self.batches = []

    def iter_batches(self) -> Iterator:
        i = 0
        while True:
            yield [i, i + 1]
            i += 2

    def process_batch(self, batch):
        self.batches.extend(batch)

    def load_state_dict(self, state: Dict):
        pass

    def state_dict(self):
        return {}


def test_learner(tmp_path: Path):
    model = TheModel()
    trainer = TheTrainer()

    learner = Learner(
        random=Random(),
        model=model,
        max_epochs=10,
        use_fp16=False,
        checkpoint_interval=1,
        trainer=trainer,
        optimizers=get_optimizers(Adam()),
        listeners=[],
        hooks=[],
    )

    TaskEnv.taskpath = tmp_path
    context = DirectoryContext(tmp_path)
    learner.instance(context=context).execute()



import logging
from typing import List, Tuple
import itertools
import pytest
import torch
from collections import defaultdict
from experimaestro import Constant
from datamaestro_text.data.ir import TextItem, create_record
from datamaestro.record import Record
from datamaestro_text.data.ir import TextItem
from xpmir.index import Index
from xpmir.learning import Random, ModuleInitMode
from xpmir.neural.dual import CosineDense, DotDense
from xpmir.letor.records import (
    PairwiseRecord,
    PairwiseRecords,
    PointwiseRecord,
    PointwiseRecords,
    ProductRecords,
)
from xpmir.text.tokenizers import Tokenizer
from xpmir.text.encoders import (
    TokenizedTextEncoderBase,
    DualTextEncoder,
    TokensRepresentationOutput,
    TokenizerOptions,
    RepresentationOutput,
)
from xpmir.text.adapters import MeanTextEncoder


class TestTokenizer(Tokenizer):
    def __init__(self):
        self.map = {}
        self._dummy_params = torch.nn.Parameter(torch.Tensor())

    def tok2id(self, tok: str) -> int:
        try:
            return self.map[tok]
        except KeyError:
            tokid = len(self.map)
            self.map[tok] = tokid
            return tokid


class RandomTokensEncoder(TokenizedTextEncoderBase[Record, TokensRepresentationOutput]):
    DIMENSION = 7
    MAX_WORDS = 100

    def __initialize__(self, options):
        super().__initialize__(options)
        self.embed = torch.nn.Embedding.from_pretrained(
            torch.randn(RandomTokensEncoder.MAX_WORDS, RandomTokensEncoder.DIMENSION)
        )
        self.tokenizer = TestTokenizer().instance()

    @property
    def dimension(self) -> int:
        return RandomTokensEncoder.DIMENSION

    @property
    def pad_tokenid(self) -> int:
        return 0

    def forward(self, records: List[Record], options=None):
        options = options or TokenizerOptions()
        tok_texts = self.tokenizer.batch_tokenize(
            [record[TextItem].text for record in records], maxlen=options.max_length, mask=True
        )
        return TokensRepresentationOutput(self.embed(tok_texts.ids), tok_texts)

    def static(self) -> bool:
        return False


class CustomIndex(Index):
    id: Constant[int] = 1

    @property
    def documentcount(self):
        return 50

    def term_df(self, term: str):
        return 1


# ---
# --- Model factories
# ---

modelfactories = []


def registermodel(method):
    modelfactories.append(
        pytest.param(method, marks=pytest.mark.dependency(name=f"model-{method}"))
    )
    return method


@registermodel
def drmm():
    """Drmm factory"""
    from xpmir.neural.interaction.drmm import Drmm
    from xpmir.neural.interaction.common import CosineSimilarity

    drmm = Drmm(
        encoder=RandomTokensEncoder(),
        index=CustomIndex(),
        similarity=CosineSimilarity(),
    )
    return drmm.instance()


@registermodel
def colbert_cos():
    """Colbert model factory"""
    from xpmir.neural.interaction.colbert import Colbert
    from xpmir.neural.interaction.common import CosineSimilarity

    return Colbert(
        encoder=RandomTokensEncoder(), similarity=CosineSimilarity()
    ).instance()


@registermodel
def dotdense():
    """Dense model factory"""
    return DotDense(
        encoder=MeanTextEncoder(encoder=RandomTokensEncoder()),
        query_encoder=MeanTextEncoder(encoder=RandomTokensEncoder()),
    ).instance()


@registermodel
def cosinedense():
    """Cosine model factory"""
    return CosineDense(
        encoder=MeanTextEncoder(encoder=RandomTokensEncoder()),
        query_encoder=MeanTextEncoder(encoder=RandomTokensEncoder()),
    ).instance()


class DummyDualTextEncoder(DualTextEncoder):
    def __init__(self):
        super().__init__()
        self.cache = defaultdict(lambda: torch.randn(1, 13))

    @property
    def dimension(self) -> int:
        return 13

    def static(self):
        return False

    def forward(self, texts: List[Tuple[str, str]]):
        return RepresentationOutput(torch.cat([self.cache[text] for text in texts]))


@registermodel
def cross_scorer():
    """Cross-scorer classifier factory"""
    from xpmir.neural.cross import CrossScorer

    return CrossScorer(max_length=100, encoder=DummyDualTextEncoder()).instance()


# ---
# --- Input factory
# ---

QUERIES = [
    create_record(text="purple cat"),
    create_record(text="yellow house"),
]
DOCUMENTS = [
    create_record(id="1", text="the cat sat on the mat"),
    create_record(id="2", text="the purple car"),
    create_record(id="3", text="my little dog"),
    create_record(id="4", text="the truck was on track"),
]


def pointwise():
    # Pointwise inputs
    inputs = PointwiseRecords()

    # Implicit order (Q0, D0) (Q1, D0) (Q0, D1) (Q1, D1)
    inputs.add(PointwiseRecord(QUERIES[0], DOCUMENTS[0], 0.0))
    inputs.add(PointwiseRecord(QUERIES[0], DOCUMENTS[0], 0.0))
    inputs.add(PointwiseRecord(QUERIES[1], DOCUMENTS[2], 0.0))
    inputs.add(PointwiseRecord(QUERIES[1], DOCUMENTS[1], 0.0))
    return inputs


def pairwise():
    # Implicit order (Q0, D0) (Q1, D0) (Q0, D1) (Q1, D1)
    inputs = PairwiseRecords()
    inputs.add(PairwiseRecord(QUERIES[0], DOCUMENTS[0], DOCUMENTS[1]))
    inputs.add(PairwiseRecord(QUERIES[1], DOCUMENTS[2], DOCUMENTS[3]))
    return inputs


def product():
    # Implicit order (Q0, D0) (Q0, D1) (Q1, D0)
    inputs = ProductRecords()
    inputs.add_topics(QUERIES[0], QUERIES[1])
    inputs.add_documents(DOCUMENTS[0], DOCUMENTS[1])

    return inputs


inputfactories = [pointwise, pairwise, product]


@pytest.mark.parametrize("modelfactory", modelfactories)
@pytest.mark.parametrize("inputfactory", inputfactories)
@pytest.mark.dependency()
def test_forward_types(modelfactory, inputfactory):
    """Test that each record type is handled"""
    model = modelfactory()
    random = Random().instance().state
    model.initialize(ModuleInitMode.RANDOM.to_options(random))

    inputs = inputfactory()

    logging.debug("%s", model(inputs, None))


@pytest.mark.parametrize("modelfactory", modelfactories)
@pytest.mark.parametrize(
    "inputfactoriescouple",
    (
        pytest.param((f1, f2), id=f"{f1.__name__}-{f2.__name__}")
        for f1, f2 in itertools.combinations(inputfactories, 2)
    ),
)
def test_forward_consistency(modelfactory, inputfactoriescouple):
    """Test that outputs are consistent between the different records types"""
    model = modelfactory()
    random = Random().instance().state
    model.initialize(ModuleInitMode.DEFAULT.to_options(random))

    outputs = []
    maps = []
    with torch.no_grad():
        for f in inputfactoriescouple:
            input = f()
            outputs.append(model(input, None))
            maps.append(
                {
                    (qr[TextItem].text, dr[TextItem].text): ix
                    for ix, (qr, dr) in enumerate(zip(input.queries, input.documents))
                }
            )

    inter = set(maps[0].keys() & maps[1].keys())
    assert len(inter) > 0, "No common query/document pair"
    for key in inter:
        s1 = outputs[0][maps[0][key]].item()
        s2 = outputs[1][maps[1][key]].item()
        assert s1 == pytest.approx(
            s2, abs=1e-6
        ), f"{s1} different from {s2} in {outputs[0]}, {outputs[1]}"



from typing import Optional, Sequence, TypeVar, Union


T = TypeVar("T")


def opt_slice(x: Optional[Sequence[T]], ix: Union[int, slice]) -> Optional[Sequence[T]]:
    if x is None:
        return None
    return x[ix]



from pathlib import Path
import json
from typing import Any, Callable, List, Optional, Dict

import numpy as np
from experimaestro import Task, Param, Meta, Annotated, pathgenerator, tqdm
from experimaestro.core.objects import Config
from datamaestro_text.data.ir import DocumentStore

from xpmir.context import Context, Hook, InitializationHook
from xpmir.documents.samplers import DocumentSampler
from xpmir.neural.generative import BeamSearchGenerationOptions
from xpmir.neural.generative.hf import T5ConditionalGenerator
from xpmir.letor import Random
from xpmir.letor.samplers import JSONLPairwiseSampleDataset
from xpmir.rankers import Retriever

from xpmir.learning import ModuleInitMode
from xpmir.learning.batchers import Batcher
from xpmir.learning.devices import DEFAULT_DEVICE, Device, DeviceInformation
from xpmir.utils.utils import batchiter, easylog, foreach

logger = easylog()


class SyntheticQueryGeneration(Task):

    model: Param[T5ConditionalGenerator]
    """The model we use to generate the queries"""

    batchsize: Meta[int] = 128
    """Batchsize when computing negatives"""

    num_qry_per_doc: Param[int] = 5
    """How many synthetic qry to generate per document"""

    sampler: Param[DocumentSampler]
    """document sampler to iterate over the corpus"""

    device: Meta[Device] = DEFAULT_DEVICE
    """The device used by the encoder"""

    batcher: Meta[Batcher] = Batcher()
    """The way to prepare batches of documents"""

    synthetic_samples: Annotated[Path, pathgenerator("synthetic.jsonl")]
    """Path to store the generated queries"""

    hooks: Param[List[Hook]] = []
    """Global learning hooks"""

    def __post_init__(self):
        super().__post_init__()
        self.generation_config = BeamSearchGenerationOptions(
            num_return_sequences=self.num_qry_per_doc,
            num_beams=self.num_qry_per_doc,
            max_new_tokens=64,
        )

    def task_outputs(self, dep: Callable[[Config], None]) -> Any:
        return dep(
            JSONLPairwiseSampleDataset(
                id=self.sampler.documents.id,
                path=self.synthetic_samples,
            )
        )

    def execute(self):
        self.device.execute(self.device_execute)

    def generate(self, batch, fp):
        generate_output = self.model.generate(
            [d.text for d in batch], self.generation_config
        )
        # length: bs*num_qry_per_doc
        queries = self.model.batch_decode(generate_output)

        # group the queries corresponds to the same document together.
        grouped_queries = [
            queries[i : i + self.num_qry_per_doc]
            for i in range(0, len(queries), self.num_qry_per_doc)
        ]
        doc_ids = [d.id for d in batch]

        for qry, doc_id in zip(grouped_queries, doc_ids):
            dict_query_doc = dict()
            dict_query_doc["queries"] = qry
            dict_query_doc["pos_ids"] = [doc_id]
            dict_query_doc["neg_ids"] = {"random": []}
            json.dump(dict_query_doc, fp)
            fp.write("\n")

    def device_execute(self, device_information: DeviceInformation):
        # Initialization hooks
        context = Context(device_information, hooks=self.hooks)
        foreach(context.hooks(InitializationHook), lambda hook: hook.before(context))

        count, iter = self.sampler()
        doc_iter = tqdm(iter, total=count, desc="Generating the synthetic queries")

        self.model.initialize(ModuleInitMode.DEFAULT.to_options())
        # put the model on the device and eval mode
        self.model.to(device_information.device).eval()
        batcher = self.batcher.initialize(self.batchsize)

        # Initialization hooks (after)
        foreach(context.hooks(InitializationHook), lambda hook: hook.after(context))

        # generate the synthetic tokens
        with self.synthetic_samples.open("wt") as fp:
            for batch in batchiter(self.batchsize, doc_iter):
                batcher.process(batch, self.generate, fp)


class JSONLNegativeGeneration(Task):
    """Add the negatives to the pairwise sampler according to the given retrievers."""

    random: Param[Optional[Random]] = None
    """The random sampler"""

    documents: Param[DocumentStore]
    """The document store where the negatives are sampling from"""

    pairwise_dataset: Param[JSONLPairwiseSampleDataset]
    """The pairwise dataset where we are going to add the negatives"""

    retrievers: Param[Dict[str, Retriever]]
    """The retrievers to retrieve the top k document wrt the query, if no
    retriever's provided, we just use the random negatives"""

    synthetic_samples: Annotated[Path, pathgenerator("synthetic_negatives.jsonl")]
    """Path to store the generated queries"""

    k: Param[int] = 100
    """The number of negatives for each algo"""

    def task_outputs(self, dep: Callable[[Config], None]) -> Any:
        return dep(
            JSONLPairwiseSampleDataset(
                id=self.documents.id,
                path=self.synthetic_samples,
            )
        )

    def execute(self):
        for retriever in self.retrievers.values():
            retriever.initialize()

        logger.info("Start to generate the negatives")

        pairwise_sample_iter = tqdm(
            self.pairwise_dataset.iter(),
            total=self.pairwise_dataset.count,
            desc="Generating negatives for the JSONL",
        )

        with self.synthetic_samples.open("wt") as fp:
            for pairwise_sample in pairwise_sample_iter:
                dict_query_doc = dict()
                query_texts = [q.text for q in pairwise_sample.topics]
                positive_ids = [pos.id for pos in pairwise_sample.positives]
                dict_query_doc["queries"] = query_texts
                dict_query_doc["pos_ids"] = positive_ids

                negatives = {}
                state = (
                    np.random.RandomState()
                    if self.random is None
                    else self.random.state
                )

                # Retrieve based on the algo
                # TODO: Make it in batch
                for (algo_name, retriever) in self.retrievers.items():
                    query_text = query_texts[state.randint(len(query_texts))]
                    scoreddocuments = retriever.retrieve(query_text)
                    ext_ids = [sd.document.id for sd in scoreddocuments]
                    filitered = [
                        ext_id for ext_id in ext_ids if ext_id not in positive_ids
                    ]
                    negatives[algo_name] = filitered
                dict_query_doc["neg_ids"] = negatives
                json.dump(dict_query_doc, fp)
                fp.write("\n")



import logging
import dataclasses
from transformers import AutoConfig, AutoTokenizer, T5ForConditionalGeneration, T5Config
from experimaestro import Param, LightweightTask
from typing import Optional, List, NamedTuple, Tuple

import torch
from torch import nn
from xpmir.learning import ModuleInitOptions, ModuleInitMode
from xpmir.text.encoders import TokenizedTexts
from xpmir.distributed import DistributableModel
from . import (
    ConditionalGenerator,
    GenerateOptions,
    BeamSearchGenerationOptions,
    StepwiseGenerator,
)


class GeneratorForwardOutput(NamedTuple):
    """The forward output of the generative retrieval"""

    logits: torch.tensor
    past_key_values: Optional[torch.tensor] = None


class FullSequenceGenerationOutput(NamedTuple):
    """The output for the generate method"""

    sequences: torch.tensor
    """The returned sequence
    shape: [bs*num_sequence, max_depth]"""

    output_mask: torch.tensor
    """A mask for the output sequences
    shape: [bs*num_sequence, max_depth]"""

    transition_scores: Optional[torch.tensor] = None
    """The condtional proba for tokens in the sequences, log, normalized
    shape: [bs * num_sequence, max_depth]"""

    all_scores: Optional[Tuple[torch.tensor]] = None
    """All the probabilities, log, normalized, tuple of length max_depth
    each tensor of the tuple has the shape of [bs * num_sequence, vs]"""

    sequence_scores: Optional[torch.tensor] = None
    """The proba for the full sequence, log
    shape: [bs * num_sequence]"""


class T5ConditionalGenerator(ConditionalGenerator, DistributableModel):

    hf_id: Param[str]
    """The HuggingFace identifier (to configure the model)"""

    def stepwise_iterator(self) -> StepwiseGenerator:
        return T5StepwiseGenerator(self)

    def __initialize__(self, options: ModuleInitOptions):
        assert options.mode != ModuleInitMode.RANDOM, "Random mode not handled (yet)"

        super().__initialize__(options)

        # Easy and hacky way to get the device
        self._dummy_params = nn.Parameter(torch.Tensor())

        self.tokenizer = AutoTokenizer.from_pretrained(self.hf_id, use_fast=True)
        self.config = AutoConfig.from_pretrained(self.hf_id)
        self.model = self.initialize_model(options)

        self.pad_token_id = self.model.config.pad_token_id
        self.decoder_start_token_id = self.model.config.decoder_start_token_id
        self.eos_token_id = self.model.config.eos_token_id

        self.encoder = self.model.get_encoder()

    def initialize_model(self, options: ModuleInitOptions):
        return T5ForConditionalGeneration(self.config)

    @property
    def device(self):
        return self._dummy_params.device

    def batch_tokenize(
        self,
        texts: List[str],
        batch_first=True,
        maxlen=None,
        mask=False,
    ) -> TokenizedTexts:
        """Tokenize the input text"""
        if maxlen is None:
            maxlen = self.tokenizer.model_max_length
        else:
            maxlen = min(maxlen, self.tokenizer.model_max_length)

        assert batch_first, "Batch first is the only option"

        r = self.tokenizer(
            list(texts),
            max_length=maxlen,
            truncation=True,
            padding=True,
            return_tensors="pt",
            return_length=True,
            return_attention_mask=mask,
        )
        return TokenizedTexts(
            None,
            r["input_ids"].to(self.device),
            r.get("length", None),
            r["attention_mask"].to(self.device) if mask else None,
            r.get("token_type_ids", None),  # if r["token_type_ids"] else None
        )

    def encode(self, texts: List[str]):
        """Returns the encoder_output and the input mask for the given text,
        which could accelerate the autoregressive generation procedure"""

        tokenized = self.batch_tokenize(texts, maxlen=512, mask=True)
        encoder_output = self.encoder(
            tokenized.ids,
            attention_mask=tokenized.mask,
            return_dict=True,
        )
        return encoder_output, tokenized.mask

    def forward(
        self,
        encoder_attention_mask,  # shape [bs, seq] with 0 or 1
        encoder_outputs,
        decoder_input_ids=None,  # if given, shape [bs]
        past_key_values=None,
    ):
        """Get the logits from the decoder"""
        bs = encoder_outputs.last_hidden_state.shape[0]

        if past_key_values is None:
            decoder_input_ids = (
                torch.ones((bs, 1), dtype=torch.long).to(self.device)
                * self.decoder_start_token_id
            )
        else:
            if decoder_input_ids is None:
                raise ValueError("decoder_input_ids of the previous layer is not given")
            else:
                decoder_input_ids = decoder_input_ids.unsqueeze(-1)

        # Do a forward pass to get the next token
        # returns three not None values:
        # past_key_values, last_hidden_state, encoder_last_hidden_state
        decoder_output = self.model(
            decoder_input_ids=decoder_input_ids,
            encoder_outputs=encoder_outputs,
            attention_mask=encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=True,
            return_dict=True,
        )
        logits = decoder_output.logits[:, -1, :]  # shape [bs, decoder_outdim+1]

        return GeneratorForwardOutput(
            logits=logits, past_key_values=decoder_output.past_key_values
        )

    def generate(
        self, inputs: List[str], options: GenerateOptions = None
    ) -> FullSequenceGenerationOutput:
        inputs = self.batch_tokenize(inputs, mask=True)
        generate_options_kwargs = dataclasses.asdict(options)
        if isinstance(options, BeamSearchGenerationOptions):
            res = self.model.generate(
                input_ids=inputs.ids,
                attention_mask=inputs.mask,
                **generate_options_kwargs,
            )
        else:
            raise NotImplementedError(
                f"Generation Options not supported for {options.__class__}"
            )

        if options.return_dict_in_generate:
            output_mask = torch.where(res.sequences != self.pad_token_id, 1, 0).to(
                self.device
            )

            if self.pad_token_id == self.decoder_start_token_id:
                output_mask[:, 0] = 1

            if not options.output_scores:
                return FullSequenceGenerationOutput(
                    sequences=res.sequences,
                    output_mask=output_mask,
                )
            else:
                # -- For the old version should be compute_transition_beam_scores
                transition_scores = self.model.compute_transition_scores(
                    res.sequences,
                    res.scores,
                    res.beam_indices,
                    normalize_logits=False,  # for bs the logits are already normalized
                )
                sequence_scores = torch.sum(transition_scores, dim=-1)
                return FullSequenceGenerationOutput(
                    sequences=res.sequences,
                    output_mask=output_mask,
                    transition_scores=transition_scores,
                    sequence_scores=sequence_scores,
                )
        else:
            output_mask = torch.where(res != self.pad_token_id, 1, 0).to(self.device)

            if self.pad_token_id == self.decoder_start_token_id:
                output_mask[:, 0] = 1

            return FullSequenceGenerationOutput(
                sequences=res,
                output_mask=output_mask,
            )

    def batch_decode(self, generate_output: FullSequenceGenerationOutput) -> List[str]:
        """Decode the sequences to meaningful texts"""
        return self.tokenizer.batch_decode(
            generate_output.sequences, skip_special_tokens=True
        )

    def distribute_models(self, update):
        self.encoder = update(self.model.get_encoder())
        self.model = update(self.model)


class T5StepwiseGenerator(StepwiseGenerator):
    def __init__(self, id_generator: ConditionalGenerator):
        super().__init__()
        # The identifier to use to generate the next step's token
        self.id_generator = id_generator

    def init(self, texts: List[str]):
        """Initialize some inner states for further iterations, and return
        the initial decoder input tokens"""
        self.encoder_output, self.attention_mask = self.id_generator.encode(texts)
        self.past_key_values = None

    def step(self, decoder_input_tokens) -> torch.Tensor:  # input shape [bs]
        """Returns the distribution over next tokens (BxV) by performing a
        stepwise iteration"""
        forward_output: GeneratorForwardOutput = self.id_generator(
            self.attention_mask,
            self.encoder_output,
            decoder_input_tokens,
            past_key_values=self.past_key_values,
        )
        self.past_key_values = forward_output.past_key_values
        return forward_output.logits


class T5ForIdentifierGeneration(T5ForConditionalGeneration):
    """T5-based identifier generation

    The class modifies T5 to use a custom vocabulary in the decoder
    """

    def __init__(self, config: T5Config, decoder_outdim: int):
        # not including the eos and pad
        self.decoder_outdim = decoder_outdim

        # modification of the config according to our needs
        config.pad_token_id = self.decoder_outdim + 1
        config.decoder_start_token_id = self.decoder_outdim + 1
        config.eos_token_id = self.decoder_outdim

        # Keep config at hand
        self.config = config

        super().__init__(self.config)

        # Modify LM head
        self.lm_head = nn.Linear(
            self.lm_head.in_features, self.decoder_outdim + 1, bias=False
        )

        # We have one more token (PAD when )
        encoder_embeddings = nn.Embedding(self.config.vocab_size, self.config.d_model)
        self.config.vocab_size = self.decoder_outdim + 1
        self.get_encoder().set_input_embeddings(encoder_embeddings)

        # Modify the decoder vocabulary
        decoder_embeddings = nn.Embedding(
            self.decoder_outdim + 2, self.config.d_model, padding_idx=decoder_outdim + 1
        )
        self.get_decoder().set_input_embeddings(decoder_embeddings)

    def forward(self, **kwargs):
        return super().forward(**kwargs)


class T5IdentifierGenerator(T5ConditionalGenerator):
    """generate the id of the token based on t5-based models"""

    decoder_outdim: Param[int] = 10
    """The decoder output dimension for the t5 model, use it to
    rebuild the lm_head and the decoder embedding, this number
    doesn't include the pad token and the eos token
    """

    def initialize_model(self, options: ModuleInitOptions):
        return T5ForIdentifierGeneration(self.config, self.decoder_outdim)


class T5ForConditionalCustomGeneration(T5ForConditionalGeneration):
    """T5-based model with custom output"""

    def __init__(self, config: T5Config, decoder_outdim: int):
        # not including the eos and pad
        self.config = config
        self.decoder_outdim = decoder_outdim
        config.decoder_start_token_id = self.decoder_outdim - 1

        super().__init__(self.config)

        # Modify LM head
        self.lm_head = nn.Linear(
            self.lm_head.in_features, self.decoder_outdim, bias=False
        )

        encoder_embeddings = nn.Embedding(self.config.vocab_size, self.config.d_model)
        self.get_encoder().set_input_embeddings(encoder_embeddings)

        # Modify the decoder vocabulary
        self.config.vocab_size = self.decoder_outdim
        decoder_embeddings = nn.Embedding(self.decoder_outdim, self.config.d_model)
        self.get_decoder().set_input_embeddings(decoder_embeddings)

    def forward(self, **kwargs):
        return super().forward(**kwargs)


class T5CustomOutputGenerator(T5ConditionalGenerator):
    """generate the id of the token based on t5-based models"""

    #: List of tokens for the output
    tokens: Param[List[str]]

    def initialize_model(self, options: ModuleInitOptions):
        return T5ForConditionalCustomGeneration(self.config, len(self.tokens))


class LoadFromT5(LightweightTask):
    """Load parameters from a T5 model"""

    t5_model: Param[T5ConditionalGenerator]
    """the target"""

    def execute(self):
        self.t5_model.initialize(ModuleInitMode.DEFAULT.to_options())

        # Load from checkpoint
        logging.info("Loading hugginface T5 from checkpoint %s", self.t5_model.hf_id)

        # Load the T5 pre-trained model
        t5_model = T5ForConditionalGeneration.from_pretrained(self.t5_model.hf_id)

        # Change the state_dict for the lm_head the decoder embedding
        state_dict = t5_model.state_dict()

        if isinstance(self.t5_model, T5IdentifierGenerator):
            # Just forget about the weights
            del state_dict["lm_head.weight"]

            # use random initialized t5 decoder
            decoder_key_names = [
                name for name in state_dict.keys() if "decoder" in name
            ]
            for name in decoder_key_names:
                del state_dict[name]
        elif isinstance(self.t5_model, T5CustomOutputGenerator):
            # Get the token embeddings from the tokenizer
            token_ids = []
            for token in self.t5_model.tokens:
                ids = self.t5_model.tokenizer.encode(token, add_special_tokens=False)
                if len(ids) != 1:
                    raise ValueError(f"Token {token} is made of {len(ids)} subtokens")
                token_ids.append(ids[0])

            # And restrict our dictionary to the possible tokens
            state_dict["lm_head.weight"] = t5_model.lm_head.weight.detach()[
                (tuple(token_ids),)
            ]
            state_dict[
                "decoder.embed_tokens.weight"
            ] = t5_model.lm_head.weight.detach()[(tuple(token_ids),)]

        logging.info("Loading state dict into the custom T5")
        self.t5_model.model.load_state_dict(state_dict, strict=False)



import torch
from torch import nn


class RbfKernelBank(nn.Module):
    def __init__(self, mus=None, sigmas=None, dim=0, requires_grad=True):
        super().__init__()
        self.mus = nn.Parameter(torch.tensor(mus), requires_grad=requires_grad)
        self.sigmas = nn.Parameter(torch.tensor(sigmas), requires_grad=requires_grad)
        self.dim = dim

    def forward(self, data):
        shape = list(data.shape)
        shape.insert(self.dim, 1)
        data = data.reshape(*shape)
        shape = [1] * len(data.shape)
        shape[self.dim] = -1
        mus, sigmas = self.mus.reshape(*shape), self.sigmas.reshape(*shape)
        adj = data - mus
        return torch.exp(-0.5 * adj * adj / sigmas / sigmas)

    def count(self):
        return self.mus.shape[0]

    @staticmethod
    def from_strs(
        mus="-0.9,-0.7,-0.5,-0.3,-0.1,0.1,0.3,0.5,0.7,0.9,1.0",
        sigmas="0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.1,0.001",
        dim=-1,
        requires_grad=True,
    ):
        mus = [float(x) for x in mus.split(",")]
        sigmas = [float(x) for x in sigmas.split(",")]
        return RbfKernelBank(mus, sigmas, dim=dim, requires_grad=requires_grad)

    @staticmethod
    def evenly_spaced(count=11, sigma=0.1, rng=(-1, 1), dim=-1, requires_grad=True):
        mus = [x.item() for x in torch.linspace(rng[0], rng[1], steps=count)]
        sigmas = [sigma for _ in mus]
        return RbfKernelBank(mus, sigmas, dim=dim, requires_grad=requires_grad)



from typing import Dict
from experimaestro import Param, field
import numpy as np
from xpmir.learning.metrics import ScalarMetric
from xpmir.letor.samplers import Sampler, SerializableIterator
from xpmir.letor.records import BaseRecords
from xpmir.utils.utils import easylog
from xpmir.learning.batchers import Batcher
from xpmir.learning.context import (
    TrainerContext,
)
from xpmir.learning.trainers import Trainer


logger = easylog()


class LossTrainer(Trainer):
    """Trainer based on a loss function

    This trainer supposes that:

    - the `sampler_iter` is initialized – and is a serializable iterator over batches
    """

    batcher: Param[Batcher] = field(default_factory=Batcher.C)
    """How to batch samples together"""

    sampler: Param[Sampler]
    """The sampler to use"""

    batch_size: Param[int] = 16
    """Number of samples per batch"""

    sampler_iter: SerializableIterator
    """The iterator over batches"""

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)

        self.sampler.initialize(random)

        self.batcher_worker = self.batcher.initialize(self.batch_size)

    def iter_batches(self) -> SerializableIterator:
        """Returns the batchwise iterator"""
        return self.sampler_iter

    def load_state_dict(self, state: Dict):
        self.sampler_iter.load_state_dict(state["sampler"])

    def state_dict(self):
        return {"sampler": self.sampler_iter.state_dict()}

    def process_batch(self, batch: BaseRecords):
        """Called by the learner to process a batch of records"""
        self.batcher_worker.process(batch, self.process_microbatch, raise_oom=True)

    def process_microbatch(self, records: BaseRecords):
        """Combines a forward and backard

        This method can be implemented by specific trainers that use the gradient.
        In that case the regularizer losses should be taken into account with
        `self.add_losses`.
        """
        # Restrict losses to this context
        with self.context.losses() as losses:
            self.train_batch(records)
            nrecords = len(records)
            total_loss = 0.0
            names = []

            for loss in losses:
                total_loss += loss.weight * loss.value
                names.append(loss.name)
                self.context.add_metric(
                    ScalarMetric(f"{loss.name}", float(loss.value.item()), nrecords)
                )

            # Reports the main metric
            if len(names) > 1:
                names.sort()
                self.context.add_metric(
                    ScalarMetric("+".join(names), float(total_loss.item()), nrecords)
                )

            self.context.state.optimizer.scale(total_loss).backward()

    def train_batch(self, records):
        """This method should report"""
        raise NotImplementedError



import sys
from typing import List
import torch
from torch import nn
from torch.functional import Tensor
from experimaestro import Config, Param
from xpmir.letor.records import (
    DocumentRecord,
    PairwiseRecord,
    PairwiseRecords,
)
from xpmir.learning.context import Loss
from xpmir.letor.trainers import TrainerContext, LossTrainer
from xpmir.utils.utils import foreach
from .samplers import DistillationPairwiseSampler, PairwiseDistillationSample
from xpmir.utils.iter import MultiprocessSerializableIterator
import numpy as np
from xpmir.rankers import LearnableScorer


class DistillationPairwiseLoss(Config, nn.Module):
    """The abstract loss for pairwise distillation"""

    weight: Param[float] = 1.0
    NAME = "?"

    def initialize(self, ranker: LearnableScorer):
        pass

    def process(
        self, student_scores: Tensor, teacher_scores: Tensor, info: TrainerContext
    ):
        loss = self.compute(student_scores, teacher_scores, info)
        info.add_loss(Loss(f"pairwise-{self.NAME}", loss, self.weight))

    def compute(
        self, student_scores: Tensor, teacher_scores: Tensor, context: TrainerContext
    ) -> torch.Tensor:
        """
        Compute the loss

        Arguments:

            student_scores: A (batch x 2) tensor
            teacher_scores: A (batch x 2) tensor
        """
        raise NotImplementedError()


class MSEDifferenceLoss(DistillationPairwiseLoss):
    """Computes the MSE between the score differences

    Compute ((student 1 - student 2) - (teacher 1 - teacher 2))**2
    """

    NAME = "delta-MSE"

    def initialize(self, ranker):
        super().initialize(ranker)
        self.loss = nn.MSELoss()

    def compute(
        self, student_scores: Tensor, teacher_scores: Tensor, info: TrainerContext
    ) -> torch.Tensor:
        return self.loss(
            student_scores[:, 1] - student_scores[:, 0],
            teacher_scores[:, 1] - teacher_scores[:, 0],
        )


class DistillationKLLoss(DistillationPairwiseLoss):
    """
    Distillation loss from: Distilling Dense Representations for Ranking using
    Tightly-Coupled Teachers https://arxiv.org/abs/2010.11386
    """

    NAME = "Distil-KL"

    def initialize(self, ranker):
        super().initialize(ranker)
        self.loss = nn.KLDivLoss(reduction="none")

    def compute(
        self, student_scores: Tensor, teacher_scores: Tensor, info: TrainerContext
    ) -> torch.Tensor:
        pos_student = student_scores[:, 0].unsqueeze(0)
        neg_student = student_scores[:, 1].unsqueeze(0)
        pos_teacher = teacher_scores[:, 0].unsqueeze(0)
        neg_teacher = teacher_scores[:, 1].unsqueeze(0)

        scores = torch.cat([pos_student, neg_student], dim=1)
        local_scores = torch.log_softmax(scores, dim=1)
        teacher_scores = torch.cat(
            [pos_teacher.unsqueeze(-1), neg_teacher.unsqueeze(-1)], dim=1
        )
        teacher_scores = torch.softmax(teacher_scores, dim=1)
        return self.loss(local_scores, teacher_scores).sum(dim=1).mean(dim=0)


class DistillationPairwiseTrainer(LossTrainer):
    """Pairwise trainer for distillation"""

    sampler: Param[DistillationPairwiseSampler]
    """The sampler"""

    lossfn: Param[DistillationPairwiseLoss]
    """The distillation pairwise batch function"""

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)
        self.lossfn.initialize(self.ranker)
        foreach(
            context.hooks(DistillationPairwiseLoss),
            lambda loss: loss.initialize(self.ranker),
        )
        self.sampler.initialize(random)
        self.sampler_iter = self.sampler.pairwise_iter()

        self.sampler_iter = MultiprocessSerializableIterator(
            self.sampler.pairwise_batch_iter(self.batch_size)
        )

    def train_batch(self, samples: List[PairwiseDistillationSample]):
        # Builds records and teacher score matrix
        teacher_scores = torch.empty(len(samples), 2)
        records = PairwiseRecords()
        for ix, sample in enumerate(samples):
            records.add(
                PairwiseRecord(
                    sample.query,
                    sample.documents[0].document,
                    sample.documents[1].document,
                )
            )
            teacher_scores[ix, 0] = sample.documents[0].score
            teacher_scores[ix, 1] = sample.documents[1].score

        # Get the next batch and compute the scores for each query/document
        scores = self.ranker(records, self.context).reshape(2, len(records)).T

        if torch.isnan(scores).any() or torch.isinf(scores).any():
            self.logger.error(
                "nan or inf relevance score detected. Aborting (pairwise distillation)."
            )
            sys.exit(1)

        # Call the losses (distillation, pairwise and pointwise)
        teacher_scores = teacher_scores.to(scores.device)
        self.lossfn.process(scores, teacher_scores, self.context)



from functools import cached_property
from typing import Iterator, List

import numpy as np
from datamaestro.record import Record
from datamaestro_text.data.conversation import (
    ConversationDataset,
    ConversationHistoryItem,
    EntryType,
)
from experimaestro import Config, Param

from xpmir.learning.base import BaseSampler, SampleIterator
from xpmir.utils.iter import RandomSerializableIterator


class DatasetConversationBase(Config):
    datasets: Param[List[ConversationDataset]]
    """The conversation datasets"""

    @cached_property
    def records(self):
        records = []
        for dataset in self.datasets:
            for conversation in dataset.__iter__():
                nodes = [
                    node
                    for node in conversation
                    if node.entry[EntryType] == EntryType.USER_QUERY
                ]
                for node in nodes:
                    records.append(
                        node.entry.update(ConversationHistoryItem(node.history()))
                    )

        return records


class DatasetConversationIterator(SampleIterator, DatasetConversationBase):
    def __iter__(self) -> Iterator[Record]:
        yield from self.records


class DatasetConversationEntrySampler(BaseSampler, DatasetConversationBase):
    """Uses a conversation dataset and topic records entries"""

    def __iter__(self) -> RandomSerializableIterator[Record]:
        return RandomSerializableIterator(self.random, self.get_iterator)

    def get_iterator(self, random: np.random.RandomState):
        return DatasetConversationEntrySamplerIterator(self, random)


class DatasetConversationEntrySamplerIterator(Iterator[Record]):
    def __init__(
        self, sampler: DatasetConversationEntrySampler, random: np.random.RandomState
    ):
        self.sampler = sampler
        self.random = random

    def __next__(self):
        if self.random is None:
            raise ValueError(
                "Random state is not initialized. Call the iterator first."
            )
        return self.sampler.records[self.random.randint(0, len(self.sampler.records))]



import sys
from experimaestro import Param
from xpmir.learning import Module
from xpmir.text.encoders import (
    TextsRepresentationOutput,
    TokenizedEncoder,
    TokenizedTexts,
    TokensRepresentationOutput,
)

from .base import HFModel


class HFEncoderBase(Module):
    """Base HuggingFace encoder"""

    model: Param[HFModel]
    """A Hugging-Face model"""

    @classmethod
    def from_pretrained_id(cls, model_id: str, **kwargs):
        """Returns a new encoder

        :param model_id: The HuggingFace Hub ID
        :param kwargs: keyword arguments passed to the model constructor
        :return: A hugging-fasce based encoder
        """
        return cls(model=HFModel.from_pretrained_id(model_id), **kwargs)

    def __initialize__(self, options):
        super().__initialize__(options)
        self.model.initialize(options)

    def static(self):
        """Embeddings from transformers are learnable"""
        return False

    @property
    def dimension(self):
        return self.model.hf_config.hidden_size

    @property
    def max_length(self):
        """Returns the maximum length that the model can process"""
        return sys.maxsize


class HFTokensEncoder(
    HFEncoderBase, TokenizedEncoder[TokenizedTexts, TokensRepresentationOutput]
):
    """HuggingFace-based tokenized"""

    def dim(self):
        return self.tokenizer.dimension

    def forward(self, tokenized: TokenizedTexts) -> TokensRepresentationOutput:
        tokenized = tokenized.to(self.model.contextual_model.device)
        y = self.model.contextual_model(
            tokenized.ids,
            attention_mask=tokenized.mask.to(self.device),
            token_type_ids=tokenized.token_type_ids,
        )
        return TokensRepresentationOutput(
            tokenized=tokenized, value=y.last_hidden_state
        )


class HFCLSEncoder(
    HFEncoderBase, TokenizedEncoder[TokenizedTexts, TextsRepresentationOutput]
):
    """Encodes a text using the [CLS] token"""

    def forward(self, tokenized: TokenizedTexts) -> TextsRepresentationOutput:
        tokenized = tokenized.to(self.device)
        y = self.model.contextual_model(
            tokenized.ids,
            attention_mask=tokenized.mask,
            token_type_ids=tokenized.token_type_ids,
        )

        # Assumes that [CLS] is the first token
        return TextsRepresentationOutput(
            tokenized=tokenized, value=y.last_hidden_state[:, 0]
        )



from dataclasses import dataclass
from typing import List
from abc import abstractmethod

import torch
from xpmir.learning.optim import Module
from xpmir.utils.utils import easylog

logger = easylog()


class StepwiseGenerator:
    """Utility class for generating one token at a time"""

    @abstractmethod
    def init(self, texts: List[str]) -> torch.Tensor:
        """Returns the distribution over the first generated tokens (BxV)
        given the texts"""
        pass

    @abstractmethod
    def step(self, token_ids: torch.LongTensor) -> torch.Tensor:
        """Returns the distribution over next tokens (BxV), given the last
        generates ones (B)"""
        pass

    @abstractmethod
    def state(self):
        """Get the current state, so we can start back to a previous generated prefix"""
        ...

    @abstractmethod
    def load_state(self, state):
        """Load a saved state"""
        ...


@dataclass
class GenerateOptions:
    """Options used during sequence generation"""

    return_dict_in_generate: bool = True
    max_new_tokens: int = 10
    """The number of new tokens to be generated"""
    output_scores: bool = True
    num_return_sequences: int = 1
    """number of returned sequences"""


@dataclass
class BeamSearchGenerationOptions(GenerateOptions):
    """Options related to the beam search of the generate method"""

    num_beams: int = 1
    """beam size"""


class ConditionalGenerator(Module):
    """Models that generate an identifier given a document or a query"""

    @abstractmethod
    def stepwise_iterator(self) -> StepwiseGenerator:
        pass

    @abstractmethod
    def generate(self, inputs: List[str], options: GenerateOptions = None):
        """Generate text given the inputs"""
        pass



import sys
from abc import ABC, abstractmethod
from dataclasses import InitVar
from typing import Dict, TypeVar, Generic

import numpy as np
import torch
from experimaestro import Config, Param

from xpmir.learning.context import Loss
from xpmir.learning.base import BaseSampler
from xpmir.learning import Module, ModuleInitMode
from xpmir.letor.trainers import LossTrainer, TrainerContext
from xpmir.utils.iter import MultiprocessSerializableIterator, SerializableIterator
from xpmir.text.encoders import RepresentationOutput

AlignementLossInput = TypeVar("AlignementLossInput")
AlignmentLossTarget = TypeVar("AlignmentLossTarget")


class AlignmentLoss(Config, ABC, Generic[AlignementLossInput, AlignmentLossTarget]):
    weight: Param[float] = 1.0
    """Weight for this loss"""

    @abstractmethod
    def __call__(
        self,
        input: AlignementLossInput,
        target: AlignmentLossTarget,
    ):
        """Computes the reconstruction loss

        :param target: a tensor of size BxD
        """
        ...


class MSEAlignmentLoss(AlignmentLoss[RepresentationOutput, RepresentationOutput]):
    """Computes the MSE between contextualized query representation and gold
    representation"""

    def __post_init__(self):
        self.mse = torch.nn.MSELoss()

    def __call__(
        self,
        input: RepresentationOutput,
        target: RepresentationOutput,
    ):
        return self.mse(input.value, target.value)


class CosineAlignmentLoss(AlignmentLoss[RepresentationOutput, RepresentationOutput]):
    """Computes the MSE between contextualized query representation and gold
    representation"""

    def __post_init__(self):
        self.loss = torch.nn.CosineEmbeddingLoss()

    def __call__(
        self,
        input: RepresentationOutput,
        target: RepresentationOutput,
    ):
        return self.loss(
            input.value, target.value, torch.Tensor([1]).to(input.value.device)
        )


class AlignmentTrainer(LossTrainer):
    """Compares two representations

    Both the representations are expected to a be in a vector space
    """

    losses: Param[Dict[str, AlignmentLoss]]
    """The loss function(s)"""

    sampler: Param[BaseSampler]
    """The pairwise sampler"""

    target_model: Param[Module]
    """Target model"""

    sampler_iter: InitVar[SerializableIterator]
    """The iterator over samples"""

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)
        self.sampler.initialize(random)
        self.sampler_iter = MultiprocessSerializableIterator(
            self.sampler.__batch_iter__(self.batch_size)
        )
        self.target_model.initialize(ModuleInitMode.DEFAULT.to_options())
        self.target_model.to(context.device_information.device)

    def train_batch(self, batch):
        # Get the next batch and compute the scores for each query/document
        output: RepresentationOutput = self.model(batch)
        with torch.no_grad():
            target: RepresentationOutput = self.target_model(batch)

        if torch.isnan(output.value).any() or torch.isinf(output.value).any():
            self.logger.error("nan or inf relevance score detected. Aborting.")
            sys.exit(1)

        for key, loss in self.losses.items():
            value = loss(output, target)
            self.context.add_loss(Loss(key, value, loss.weight))



import logging
import json
from pathlib import Path
from typing import Dict, Iterator, List
from datamaestro_text.data.ir import Adhoc
from experimaestro import Param, pathgenerator, Annotated
import numpy as np
from xpmir.utils.utils import easylog, foreach
from xpmir.evaluation import evaluate
from xpmir.learning.context import TrainState, TrainerContext, ValidationHook
from xpmir.rankers import (
    Retriever,
)
from xpmir.learning.learner import LearnerListener, Learner, LearnerListenerStatus
from xpmir.learning.optim import ModuleLoader

logger = easylog()


class ValidationModuleLoader(ModuleLoader):
    """Specializes the validation listener"""

    listener: Param["LearnerListener"]
    """The listener"""

    key: Param[str]
    """The key for this listener"""


class ValidationListener(LearnerListener):
    """Learning validation early-stopping

    Computes a validation metric and stores the best result. If early_stop is
    set (> 0), then it signals to the learner that the learning process can
    stop.
    """

    metrics: Param[Dict[str, bool]] = {"map": True}
    """Dictionary whose keys are the metrics to record, and boolean
            values whether the best performance checkpoint should be kept for
            the associated metric ([parseable by ir-measures](https://ir-measur.es/))"""

    dataset: Param[Adhoc]
    """The dataset to use"""

    retriever: Param[Retriever]
    """The retriever for validation"""

    warmup: Param[int] = -1
    """How many epochs before actually computing the metric"""

    bestpath: Annotated[Path, pathgenerator("best")]
    """Path to the best checkpoints"""

    info: Annotated[Path, pathgenerator("info.json")]
    """Path to the JSON file that contains the metric values at each epoch"""

    validation_interval: Param[int] = 1
    """Epochs between each validation"""

    early_stop: Param[int] = 0
    """Number of epochs without improvement after which we stop learning.
    Should be a multiple of validation_interval or 0 (no early stopping)"""

    hooks: Param[List[ValidationHook]] = []
    """The list of the hooks during the validation"""

    def __validate__(self):
        assert (
            self.early_stop % self.validation_interval == 0
        ), "Early stop should be a multiple of the validation interval"

    def initialize(self, learner: Learner, context: TrainerContext):
        super().initialize(learner, context)

        self.retriever.initialize()
        self.bestpath.mkdir(exist_ok=True, parents=True)

        # Checkpoint start
        try:
            with self.info.open("rt") as fp:
                self.top = json.load(fp)  # type: Dict[str, Dict[str, float]]
        except Exception:
            self.top = {}

    def update_metrics(self, metrics: Dict[str, float]):
        if self.top:
            # Just use another key
            for metric in self.metrics.keys():
                metrics[f"{self.id}/final/{metric}"] = self.top[metric]["value"]

    def monitored(self) -> Iterator[str]:
        return [key for key, monitored in self.metrics.items() if monitored]

    def init_task(self, learner: "Learner", dep):
        return {
            key: dep(
                ValidationModuleLoader(
                    value=learner.model,
                    listener=self,
                    key=key,
                    path=self.bestpath / key / TrainState.MODEL_PATH,
                )
            )
            for key, store in self.metrics.items()
            if store
        }

    def should_stop(self, epoch=0):
        if self.early_stop > 0 and self.top:
            epochs_since_imp = (epoch or self.context.epoch) - max(
                info["epoch"] for key, info in self.top.items() if self.metrics[key]
            )
            if epochs_since_imp >= self.early_stop:
                return LearnerListenerStatus.STOP

        return LearnerListenerStatus.DONT_STOP

    def __call__(self, state: TrainState):
        # Check that we did not stop earlier (when loading from checkpoint / if other
        # listeners have not stopped yet)
        if self.should_stop(state.epoch - 1) == LearnerListenerStatus.STOP:
            return LearnerListenerStatus.STOP

        foreach(
            self.hooks,
            lambda hook: hook.before(self.context),
        )

        if state.epoch % self.validation_interval == 0:
            # Compute validation metrics
            means, details = evaluate(
                self.retriever, self.dataset, list(self.metrics.keys()), True
            )

            for metric, keep in self.metrics.items():
                value = means[metric]

                self.context.writer.add_scalar(
                    f"{self.id}/{metric}/mean", value, state.step
                )

                self.context.writer.add_histogram(
                    f"{self.id}/{metric}",
                    np.array(list(details[metric].values()), dtype=np.float32),
                    state.step,
                )

                # Update the top validation
                if state.epoch >= self.warmup:
                    topstate = self.top.get(metric, None)
                    if topstate is None or value > topstate["value"]:
                        # Save the new top JSON
                        self.top[metric] = {"value": value, "epoch": self.context.epoch}

                        # Copy in corresponding directory
                        if keep:
                            logging.info(
                                f"Saving the checkpoint {state.epoch}"
                                f" for metric {metric}"
                            )
                            self.context.copy(self.bestpath / metric)

            # Update information
            with self.info.open("wt") as fp:
                json.dump(self.top, fp)

        foreach(
            self.hooks,
            lambda hook: hook.after(self.context),
        )

        # Early stopping?
        return self.should_stop()



from typing import Callable, List
from experimaestro import Config, Param
import torch.nn as nn
import torch
from abc import ABC, abstractmethod
from xpmir.context import InitializationHook
from xpmir.learning.devices import ComputationContext, DistributedDeviceInformation
from xpmir.utils.utils import easylog

logger = easylog()


class DistributableModel(Config, ABC):
    """A model that can be distributed over GPUs

    Subclasses must implement :py:meth:`distribute_models`
    """

    @abstractmethod
    def distribute_models(self, update: Callable[[nn.Module], nn.Module]):
        """This method is called with an `update` parameter that should be used
        to update all the torch modules that we need to distribute on GPUs"""
        ...


class DistributedHook(InitializationHook):
    """Hook to distribute the model processing

    When in multiprocessing/multidevice, use `torch.nn.parallel.DistributedDataParallel`
    ,otherwise use `torch.nn.DataParallel`.
    """

    models: Param[List[DistributableModel]]
    """The model to distribute over GPUs"""

    def after(self, state: ComputationContext):
        for model in self.models:
            model.distribute_models(lambda model: self.update(state, model))

    def update(self, state: ComputationContext, model: nn.Module) -> nn.Module:
        info = state.device_information
        if isinstance(info, DistributedDeviceInformation):
            logger.info("Using a distributed model with rank=%d", info.rank)
            return nn.parallel.DistributedDataParallel(model, device_ids=[info.rank])
        else:
            if not isinstance(model, nn.DataParallel):
                n_gpus = torch.cuda.device_count()
                if n_gpus > 1:
                    logger.info(
                        "Setting up DataParallel on %d GPUs for model %s",
                        n_gpus,
                        str(model.__class__.__qualname__),
                    )
                    return DataParallel(model)
                else:
                    logger.warning("Only one GPU detected, not using data parallel")

        return model


class DataParallel(torch.nn.DataParallel):
    """Subclasses DataParallel for serialization

    Removes the `.module` prefix introduced by DataParallel by:

    - (loading) Registering a load state hook
    - (saving) Redefining the state dict to skip the module prefix
    """

    module: nn.Module
    """The model to put on multi dataset."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._register_load_state_dict_pre_hook(DataParallel.load_state_dict_hook)

    @staticmethod
    def load_state_dict_hook(
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        replaced_keys = [key for key in state_dict if key.startswith(prefix)]
        offset = len(prefix)
        for key in replaced_keys:
            new_key = f"{prefix}module.{key[offset:]}"
            state_dict[new_key] = state_dict.pop(key)

    def state_dict(self, **kwargs):
        return self.module.state_dict(**kwargs)



from typing import List

from datamaestro.record import Record
from experimaestro import Param
from datamaestro_text.data.ir import TextItem
from xpmir.utils.convert import Converter

from .encoders import InputType, RepresentationOutput, TokenizedTextEncoderBase


class MeanTextEncoder(TokenizedTextEncoderBase[InputType, RepresentationOutput]):
    """Returns the mean of the word embeddings"""

    encoder: Param[TokenizedTextEncoderBase[InputType, RepresentationOutput]]

    def __initialize__(self, options):
        self.encoder.__initialize__(options)

    def static(self):
        return self.encoder.static()

    @property
    def dimension(self):
        return self.encoder.dimension

    def forward(self, texts: List[InputType], options=None) -> RepresentationOutput:
        # emb_texts = self.encoder(texts, options=options)
        emb_texts = self.encoder(texts, options=options)
        # Computes the mean over the time dimension (vocab output is batch x time x dim)
        emb_texts.value = emb_texts.value.mean(1)
        return emb_texts


class TopicTextConverter(Converter[Record, str]):
    """Extracts the text from a topic"""

    def __call__(self, input: Record) -> str:
        return input[TextItem].text



from experimaestro import ObjectStore
import torch.nn as nn
from itertools import chain
from xpmir.learning.optim import Module, ModuleList


class MyModule(Module):
    def __post_init__(self) -> None:
        self.linear = nn.Linear(2, 3)


def test_module_list():
    a = MyModule()
    b = MyModule()
    container = ModuleList(sub_modules=[a, b])

    store = ObjectStore()
    container = container.instance(objects=store)
    a = a.instance(objects=store)
    b = b.instance(objects=store)
    assert set(container.parameters()) == set(chain(a.parameters(), b.parameters()))



import os
from pathlib import Path
from typing import Optional, Union, Dict
import shutil
from experimaestro.huggingface import ExperimaestroHFHub
from experimaestro import Config
from xpmir.neural.dual import DotDense
from xpmir.neural.huggingface import HFCrossScorer
from xpmir.utils.utils import easylog
from xpmir.learning.optim import ModuleLoader
import importlib

logger = easylog()


def get_class(name: str):
    module_name, class_name = name.split(":")
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


class XPMIRHFHub(ExperimaestroHFHub):
    def __init__(
        self,
        config: Config,
        variant: Optional[str] = None,
        readme: Optional[str] = None,
        tb_logs: Optional[Dict[str, Path]] = None,
    ):
        super().__init__(config, variant)
        self.readme = readme
        self.tb_logs = tb_logs

    def _save_pretrained(self, save_directory: Union[str, Path]):
        save_directory = Path(save_directory)
        super()._save_pretrained(save_directory)
        if self.readme:
            (save_directory / "README.md").write_text(self.readme)

        if self.tb_logs:
            runs_dir = save_directory / "runs"
            runs_dir.mkdir()
            for key, path in self.tb_logs.items():
                shutil.copytree(path, runs_dir / key)


class AutoModel:
    @staticmethod
    def load_from_hf_hub(
        hf_id_or_folder: str, variant: Optional[str] = None, as_instance: bool = False
    ):
        """Loads from hugging face hub or from a folder"""
        local_files_only = os.environ("HF_HUB_OFFLINE")
        data = XPMIRHFHub.from_pretrained(
            hf_id_or_folder,
            variant=variant,
            as_instance=as_instance,
            local_files_only=local_files_only,
        )

        if isinstance(data, ModuleLoader):
            model, init_tasks = data.value, [data]
        else:
            raise Exception(f"Cannot handle data of type {type(data)}")

        if as_instance:
            for init_task in init_tasks:
                init_task.execute()
            return model
        return model, init_tasks

    @staticmethod
    def push_to_hf_hub(config: Config, *args, variant=None, readme=None, **kwargs):
        """Push to HuggingFace Hub

        See ModelHubMixin.push_to_hub for the other arguments
        """
        return XPMIRHFHub(config, variant=variant, readme=readme).push_to_hub(
            *args, **kwargs
        )

    @staticmethod
    def sentence_scorer(hf_id: str):
        """Loads from hugging face hub using a sentence transformer"""
        try:
            from sentence_transformers import SentenceTransformer
        except Exception:
            logger.error(
                "Sentence transformer is not installed:"
                "pip install -U sentence_transformers"
            )
            raise

        encoder = SentenceTransformer(hf_id)
        return DotDense(encoder=encoder)

    @staticmethod
    def cross_encoder_model(hf_id: str, max_length: int = 512):
        """Loads from huggingface hub in to a form of a cross-scorer, it returns
        a sentence_transformer model for cross encoder"""

        scorer = HFCrossScorer(hf_id=hf_id, max_length=max_length)
        return scorer



import logging
from functools import cached_property


def easylog():
    """
    Returns a logger with the caller's __name__
    """
    import inspect

    try:
        frame = inspect.stack()[1]  # caller
        module = inspect.getmodule(frame[0])
        return logging.getLogger(module.__name__)
    except IndexError:
        return logging.getLogger("UNKNOWN")


class EasyLogger:
    @cached_property
    def logger(self):
        clsdict = self.__class__.__dict__

        logger = clsdict.get("__LOGGER__", None)
        if logger is None:
            logger = logging.getLogger(self.__class__.__qualname__)
            self.__class__.__LOGGER__ = logger

        return logger


class LazyJoin:
    """Lazy join of an iterator"""

    def __init__(self, glue: str, iterator):
        self.glue = glue
        self.iterator = iterator

    def __str__(self):
        return self.glue.join(str(x) for x in self.iterator)



from typing import List, Optional, Generic
from experimaestro import Config, Param
from datamaestro_text.data.ir import TextItem
import torch.nn as nn
import torch
from xpmir.learning import ModuleInitOptions
from xpmir.distributed import DistributableModel
from xpmir.text.huggingface import (
    OneHotHuggingFaceEncoder,
    TransformerTokensEncoderWithMLMOutput,
)
from xpmir.text import TokenizerOptions
from xpmir.text.huggingface import HFTokenizerBase
from xpmir.text.encoders import (
    TextEncoder,
    TextEncoderBase,
    InputType as EncoderInputType,
    TextsRepresentationOutput,
)
from xpmir.neural.dual import DotDense, ScheduledFlopsRegularizer
from xpmir.text.huggingface.base import HFMaskedLanguageModel
from xpmir.utils.utils import easylog

logger = easylog()


class Aggregation(Config):
    """The aggregation function for Splade"""

    def get_output_module(self, linear: nn.Module) -> nn.Module:
        return AggregationModule(linear, self)


class MaxAggregation(Aggregation):
    """Aggregate using a max"""

    def __call__(self, logits, mask):
        # Get the maximum (masking the values)
        values, _ = torch.max(
            torch.relu(logits) * mask.to(logits.device).unsqueeze(-1),
            dim=1,
        )

        # Computes log(1+x)
        return torch.log1p(values.clamp(min=0))


class SumAggregation(Aggregation):
    """Aggregate using a sum"""

    def __call__(self, logits, mask):
        return torch.sum(
            torch.log1p(torch.relu(logits) * mask.to(logits.device).unsqueeze(-1)),
            dim=1,
        )


class AggregationModule(nn.Module):
    def __init__(self, linear: nn.Linear, aggregation: Aggregation):
        super().__init__()
        self.linear = linear
        self.aggregation = aggregation

    def forward(self, input: torch.Tensor, mask: torch.Tensor):
        return self.aggregation(self.linear(input), mask)


class SpladeTextEncoderModel(nn.Module):
    def __init__(
        self, encoder: TransformerTokensEncoderWithMLMOutput, aggregation: Aggregation
    ):
        super().__init__()
        self.encoder = encoder
        self.aggregation = aggregation

    def forward(self, tokenized):
        # We stock all the outputs in order to get the embedding matrix
        # Here as the automodel is not the same as the normal AutoModel,
        # So here the output has the attribute logits, the w_ij in the paper
        # which is of shape (bs, len(texts), vocab_size)
        out = self.encoder(tokenized, all_outputs=True)
        out = self.aggregation(out.logits, tokenized.mask)
        return out


class SpladeTextEncoder(TextEncoder, DistributableModel):
    """Splade model

    It is only a text encoder since the we use `xpmir.neural.dual.DotDense`
    as the scorer class
    """

    encoder: Param[TransformerTokensEncoderWithMLMOutput]
    """The encoder from Hugging Face"""

    aggregation: Param[Aggregation]
    """How to aggregate the vectors"""

    maxlen: Param[Optional[int]] = None
    """Max length for texts"""

    def __initialize__(self, options: ModuleInitOptions):
        self.encoder.initialize(options)
        self.model = SpladeTextEncoderModel(self.encoder, self.aggregation)

    def forward(self, texts: List[str]) -> torch.Tensor:
        """Returns a batch x vocab tensor"""
        if not isinstance(texts[0], str):
            texts = [text[TextItem].text for text in texts]
        tokenized = self.encoder.batch_tokenize(texts, mask=True, maxlen=self.maxlen)
        out = self.model(tokenized)
        return TextsRepresentationOutput(out, tokenized)

    @property
    def dimension(self):
        return self.encoder.model.config.vocab_size

    def static(self):
        return False

    def distribute_models(self, update):
        self.model = update(self.model)


class IdentityWithBias(nn.Identity):
    def __init__(self):
        # So that set_output_embeddings is happy
        super().__init__()
        self.bias = None


class SpladeTextEncoderV2(
    TextEncoderBase[EncoderInputType, TextsRepresentationOutput],
    DistributableModel,
    Generic[EncoderInputType],
):
    # TODO: use "SpladeTextEncoder" identifier until
    # https://github.com/experimaestro/experimaestro-python/issues/56 is fixed
    __xpmid__ = str(SpladeTextEncoder.__getxpmtype__().identifier)

    """Splade model text encoder (V2)

    It is only a text encoder since the we use `xpmir.neural.dual.DotDense`
    as the scorer class. Compared to V1, it uses the new text HF encoder abstractions.
    """

    tokenizer: Param[HFTokenizerBase[EncoderInputType]]
    """The tokenizer from Hugging Face"""

    encoder: Param[HFMaskedLanguageModel]
    """The encoder from Hugging Face"""

    aggregation: Param[Aggregation]
    """How to aggregate the vectors"""

    maxlen: Param[Optional[int]] = None
    """Max length for texts"""

    def __initialize__(self, options: ModuleInitOptions):
        self.encoder.initialize(options)
        self.tokenizer.initialize(options)

        # Adds the aggregation head right away - this could allows
        # optimization e.g. for the Max aggregation method
        output_embeddings = self.encoder.model.get_output_embeddings()
        assert isinstance(
            output_embeddings, nn.Linear
        ), f"Cannot handle output embeddings of class {output_embeddings.__cls__}"
        self.encoder.model.set_output_embeddings(IdentityWithBias())

        self.aggregation = self.aggregation.get_output_module(output_embeddings)

    def forward(self, texts: EncoderInputType) -> TextsRepresentationOutput:
        """Returns a batch x vocab tensor"""
        tokenized = self.tokenizer.tokenize(
            texts, options=TokenizerOptions(self.maxlen)
        )

        value = self.aggregation(self.encoder(tokenized).logits, tokenized.mask)
        return TextsRepresentationOutput(value, tokenized)

    @property
    def dimension(self):
        return self.encoder.model.config.vocab_size

    def static(self):
        return False

    def distribute_models(self, update):
        self.encoder = update(self.encoder)


def _splade(
    lambda_q: float,
    lambda_d: float,
    aggregation: Aggregation,
    lambda_warmup_steps: int = 0,
    hf_id: str = "distilbert-base-uncased",
):
    # Unlike the cross-encoder, here the encoder returns the whole last layer
    # In the paper we use the DistilBERT-based as the checkpoint
    encoder = TransformerTokensEncoderWithMLMOutput(model_id=hf_id, trainable=True)

    # make use the output of the BERT and do an aggregation
    doc_encoder = SpladeTextEncoder(
        aggregation=aggregation, encoder=encoder, maxlen=200
    )
    query_encoder = SpladeTextEncoder(
        aggregation=aggregation, encoder=encoder, maxlen=30
    )

    return DotDense(
        encoder=doc_encoder, query_encoder=query_encoder
    ), ScheduledFlopsRegularizer(
        lambda_q=lambda_q,
        lambda_d=lambda_d,
        lambda_warmup_steps=lambda_warmup_steps,
    )


def _splade_doc(
    lambda_q: float,
    lambda_d: float,
    aggregation: Aggregation,
    lambda_warmup_steps: int = 0,
    hf_id: str = "distilbert-base-uncased",
):
    # Unlike the cross-encoder, here the encoder returns the whole last layer
    # The doc_encoder is the traditional one, and the query encoder return a vector
    # contains only 0 and 1
    # In the paper we use the DistilBERT-based as the checkpoint
    encoder = TransformerTokensEncoderWithMLMOutput(model_id=hf_id, trainable=True)

    # make use the output of the BERT and do an aggregation
    doc_encoder = SpladeTextEncoder(
        aggregation=aggregation, encoder=encoder, maxlen=256
    )

    query_encoder = OneHotHuggingFaceEncoder(model_id=hf_id, maxlen=30)

    return DotDense(
        encoder=doc_encoder, query_encoder=query_encoder
    ), ScheduledFlopsRegularizer(
        lambda_q=lambda_q,
        lambda_d=lambda_d,
        lambda_warmup_steps=lambda_warmup_steps,
    )


def spladeV1(
    lambda_q: float,
    lambda_d: float,
    lambda_warmup_steps: int = 0,
    hf_id: str = "distilbert-base-uncased",
):
    """Returns the Splade architecture"""
    return _splade(lambda_q, lambda_d, SumAggregation(), lambda_warmup_steps, hf_id)


def spladeV2_max(
    lambda_q: float,
    lambda_d: float,
    lambda_warmup_steps: int = 0,
    hf_id: str = "distilbert-base-uncased",
):
    """Returns the Splade-max architecture

    SPLADE v2: Sparse Lexical and Expansion Model for Information Retrieval
    (arXiv:2109.10086)
    """
    return _splade(lambda_q, lambda_d, MaxAggregation(), lambda_warmup_steps, hf_id)


def spladeV2_doc(
    lambda_q: float,
    lambda_d: float,
    lambda_warmup_steps: int = 0,
    hf_id: str = "distilbert-base-uncased",
):
    """Returns the Splade-doc architecture

    SPLADE v2: Sparse Lexical and Expansion Model for Information Retrieval
    (arXiv:2109.10086)
    """
    return _splade_doc(lambda_q, lambda_d, MaxAggregation(), lambda_warmup_steps, hf_id)



from dataclasses import InitVar
import math
import sys
from typing import Iterator, Any
import torch
from torch import nn
from torch.functional import Tensor
import torch.nn.functional as F
from experimaestro import Config, Param
from xpmir.learning.context import Loss
from xpmir.learning.metrics import ScalarMetric
from xpmir.letor.records import (
    PairwiseRecord,
    PairwiseRecordWithTarget,
    PairwiseRecords,
    PairwiseRecordsWithTarget,
)
from xpmir.letor.samplers import PairwiseSampler, SerializableIterator
from xpmir.letor.trainers import TrainerContext, LossTrainer
import numpy as np
from xpmir.rankers import LearnableScorer, ScorerOutputType
from xpmir.utils.utils import foreach
from xpmir.utils.iter import MultiprocessSerializableIterator
from xpmir.utils.utils import EasyLogger
from xpmir.learning.losses import bce_with_logits_loss


class PairwiseLoss(Config, nn.Module):
    """Base class for any pairwise loss"""

    NAME = "?"

    weight: Param[float] = 1.0
    """The weight :math:`w` with which the loss is multiplied (useful when
    combining with other ones)"""

    def initialize(self, ranker: LearnableScorer):
        pass

    def process(self, scores: Tensor, context: TrainerContext):
        value = self.compute(scores, context)
        context.add_loss(Loss(f"pair-{self.NAME}", value, self.weight))

    def compute(self, scores: Tensor, info: TrainerContext) -> Tensor:

        """Computes the loss

        :param scores: A (batch x 2) tensor (positive/negative)
        :param info: the trainer context
        :return: a torch scalar
        """
        raise NotImplementedError()


class CrossEntropyLoss(PairwiseLoss):
    r"""Cross-Entropy Loss

    Computes the cross-entropy loss

    Classification loss (relevant vs non-relevant) where the logit
    is equal to the difference between the relevant and the non relevant
    document (or equivalently, softmax then mean log probability of relevant documents)
    Reference: C. Burges et al., “Learning to rank using gradient descent,” 2005.

    *warning*: this loss assumes the score returned by the scorer is a logit

    .. math::

        \frac{w}{N} \sum_{(s^+,s-)} \log \frac{\exp(s^+)}{\exp(s^+)+\exp(s^-)}
    """
    NAME = "cross-entropy"

    def compute(self, rel_scores_by_record, info: TrainerContext):
        target = (
            torch.zeros(rel_scores_by_record.shape[0])
            .long()
            .to(rel_scores_by_record.device)
        )
        return F.cross_entropy(rel_scores_by_record, target, reduction="mean")


class HingeLoss(PairwiseLoss):
    r"""Hinge (or max-margin) loss

    .. math::

       \frac{w}{N} \sum_{(s^+,s-)} \max(0, m - (s^+ - s^-))

    """

    NAME = "hinge"

    margin: Param[float] = 1.0
    """The margin for the Hinge loss"""

    def compute(self, rel_scores_by_record, info: TrainerContext):
        return F.relu(
            self.margin - rel_scores_by_record[:, 0] + rel_scores_by_record[:, 1]
        ).mean()


class PointwiseCrossEntropyLoss(PairwiseLoss, EasyLogger):
    r"""Point-wise cross-entropy loss

    This is a point-wise loss adapted as a pairwise one.

    This loss adapts to the ranker output type:

    - If real, uses a BCELossWithLogits (sigmoid transformation)
    - If probability, uses the BCELoss
    - If log probability, uses a BCEWithLogLoss

    .. math::

        \frac{w}{2N} \sum_{(s^+,s-)} \log \frac{\exp(s^+)}{\exp(s^+)+\exp(s^-)}
        + \log \frac{\exp(s^-)}{\exp(s^+)+\exp(s^-)}

    """

    NAME = "pointwise-cross-entropy"

    def initialize(self, ranker: LearnableScorer):
        super().initialize(ranker)
        self.rankerOutputType = ranker.outputType
        if ranker.outputType == ScorerOutputType.REAL:
            self.logger.info("Ranker outputs logits: using BCEWithLogitsLoss")
            self.loss = nn.BCEWithLogitsLoss()
        elif ranker.outputType == ScorerOutputType.PROBABILITY:
            self.logger.info("Ranker outputs probabilities: using BCELoss")
            self.loss = nn.BCELoss()
        elif ranker.outputType == ScorerOutputType.LOG_PROBABILITY:
            self.logger.info("Ranker outputs probabilities: using BCEWithLogLoss")
            self.loss = bce_with_logits_loss
        else:
            raise Exception("Not implemented")

    def compute(self, rel_scores_by_record, info: TrainerContext):
        device = rel_scores_by_record.device
        dim = rel_scores_by_record.shape[0]
        target = torch.cat(
            (torch.ones(dim, device=device), torch.zeros(dim, device=device))
        )
        return self.loss(rel_scores_by_record.T.flatten(), target)


class PairwiseTrainer(LossTrainer):
    """Pairwise trainer uses samples of the form (query, positive, negative)"""

    lossfn: Param[PairwiseLoss]
    """The loss function"""

    sampler: Param[PairwiseSampler]
    """The pairwise sampler"""

    sampler_iter: InitVar[SerializableIterator[PairwiseRecord, Any]]

    def initialize(
        self,
        random: np.random.RandomState,
        context: TrainerContext,
    ):
        super().initialize(random, context)
        self.lossfn.initialize(self.ranker)
        foreach(context.hooks(PairwiseLoss), lambda loss: loss.initialize(self.ranker))
        self.sampler.initialize(random)
        self.sampler_iter = MultiprocessSerializableIterator(
            self.sampler.pairwise_batch_iter(self.batch_size)
        )

    def train_batch(self, records: PairwiseRecords):
        # Get the next batch and compute the scores for each query/document
        rel_scores = self.ranker(records, self.context)

        if torch.isnan(rel_scores).any() or torch.isinf(rel_scores).any():
            self.logger.error("nan or inf relevance score detected. Aborting.")
            sys.exit(1)

        # Reshape to get the pairs and compute the loss
        pairwise_scores = rel_scores.reshape(2, len(records)).T
        self.lossfn.process(pairwise_scores, self.context)

        self.context.add_metric(
            ScalarMetric(
                "accuracy", float(self.acc(pairwise_scores).item()), len(rel_scores)
            )
        )

    def acc(self, scores_by_record) -> Tensor:
        with torch.no_grad():
            return (
                scores_by_record[:, 0] > scores_by_record[:, 1]
            ).sum().float() / len(scores_by_record)


class PairwiseLossWithTarget(Config):
    NAME = "?"
    weight: Param[float] = 1.0

    def initialize(self, ranker: LearnableScorer):
        pass

    def process(self, scores: Tensor, targets: Tensor, context: TrainerContext):
        value = self.compute(scores, targets, context)
        context.add_loss(Loss(f"duo-{self.NAME}", value, self.weight))


class PairwiseLossWithTarget(PairwiseLossWithTarget):
    NAME = "logproba"

    def initialize(self, ranker: LearnableScorer):
        self.loss = {
            ScorerOutputType.REAL: nn.BCEWithLogitsLoss,
            ScorerOutputType.LOG_PROBABILITY: None,
            ScorerOutputType.PROBABILITY: nn.BCELoss,
        }[ranker.outputType]()

    def compute(self, scores: Tensor, targets: Tensor, context: TrainerContext):
        return self.loss(scores, targets)


class DuoPairwiseTrainer(LossTrainer):
    """The pairwise trainer for duobert. The iter_batch method
    can be the same as the pairwiseTrainer
    """

    lossfn: Param[PairwiseLossWithTarget]
    """The loss function"""

    sampler: Param[PairwiseSampler]
    """The pairwise sampler"""

    sampler_iter: InitVar[SerializableIterator[PairwiseRecord, Any]]

    def initialize(self, random: np.random.RandomState, context: TrainerContext):
        super().initialize(random, context)
        self.lossfn.initialize(self.ranker)

        self.score_threshold = {
            ScorerOutputType.LOG_PROBABILITY: math.log(0.5),
            ScorerOutputType.PROBABILITY: 0.5,
            ScorerOutputType.REAL: 0,
        }[self.ranker.outputType]
        foreach(context.hooks(PairwiseLoss), lambda loss: loss.initialize(self.ranker))
        self.sampler.initialize(random)
        self.sampler_iter = self.sampler.pairwise_iter()

    def iter_batches(self) -> Iterator[PairwiseRecordsWithTarget]:
        while True:
            batch = PairwiseRecordsWithTarget()
            for _, record in zip(range(self.batch_size), self.sampler_iter):
                # randomly swap the first and second document
                if self.random.random() < 0.5:
                    batch.add(
                        PairwiseRecordWithTarget(
                            record.query, record.positive, record.negative, 1
                        )
                    )
                else:
                    batch.add(
                        PairwiseRecordWithTarget(
                            record.query, record.negative, record.positive, 0
                        )
                    )
            yield batch

    def train_batch(self, records: PairwiseRecords):
        # Get the next batch and compute the scores for each query/document
        # forward pass
        rel_scores = self.ranker(records, self.context)  # shape: (bs)
        targets = torch.Tensor(records.get_target()).to(rel_scores.device)

        if torch.isnan(rel_scores).any() or torch.isinf(rel_scores).any():
            self.logger.error("nan or inf relevance score detected. Aborting.")
            sys.exit(1)

        # Reshape to get the pairs and compute the loss
        self.lossfn.process(
            rel_scores,
            torch.Tensor(targets),
            self.context,
        )

        self.context.add_metric(
            ScalarMetric(
                "accuracy",
                self.acc(
                    rel_scores,
                    torch.Tensor(targets),
                ).item(),
                len(rel_scores),
            )
        )

    def acc(self, scores_by_record, target) -> Tensor:
        with torch.no_grad():
            positives = scores_by_record > self.score_threshold
            return (positives == target).sum() / len(positives)



from experimaestro import Param, Config


class Model(Config):
    """Base class for standard IR models"""

    pass


class BM25(Model):
    """BM-25 model definition"""

    k1: Param[float] = 0.9
    b: Param[float] = 0.4


class QLDirichlet(Model):
    """Query likelihood (Dirichlet smoothing) model definition"""

    mu: Param[float] = 1000
